"""Module realized error estimation after minimization procedure by simplex.

Functions
---------
    - error_estimation_simplex

"""
import numpy


def error_estimation_simplex(vertex_vector_h, vertex_chi_sq_h, func):
    """
    Error estimation.

    Calculations according to
    ANALYTICAL CHEMISTRY, VOL. 60, NO. 8, APRIL 15, 1988

    notations
    ---------
        theta_i = vertex_vector[i, :]
        chi_sq_i = vertex_chi_sq[i]
    """
    # print("\nvertex_vector")
    # print(vertex_vector_h)
    # print("\nvertex_chi_sq")
    # print(vertex_chi_sq_h)

    # temporary solution
    k, hh = vertex_vector_h.shape  # hh = k-1
    theta_0 = vertex_vector_h[0, :]
    m_q = numpy.zeros((k-1, k-1))
    vertex_vector = numpy.zeros(vertex_vector_h.shape, dtype=float)
    vertex_vector[0, :] = theta_0
    max_radius = numpy.zeros(k-1, dtype=float)
    for i in range(1, k):
        theta_i = vertex_vector_h[i, :]
        rand_radius = numpy.abs(theta_i-theta_0)
        max_radius = numpy.max(numpy.vstack([max_radius, rand_radius]), axis=0)
    # print("max_radius ", max_radius)
    for i in range(1, k):
        radius_h = numpy.zeros(k-1, dtype=float)
        radius_h[i-1] = max_radius[i-1]
        vertex_vector[i, :] = theta_0+radius_h

    l_chi_sq = []
    for i in range(0, k):
        theta_i = vertex_vector_h[i, :]
        chi_sq = func(theta_i)
        l_chi_sq.append(chi_sq)
    vertex_chi_sq = numpy.array(l_chi_sq, dtype=float)

    # print("hh, k: ", hh, k)
    # print("theta_0: ", theta_0)
    chi_sq_0 = vertex_chi_sq[0]
    # print("chi_sq_0: ", chi_sq_0)
    v_a = numpy.zeros(k-1)
    m_b = numpy.zeros((k-1, k-1))
    m_q = numpy.zeros((k-1, k-1))
    m_chi_sq_0i = numpy.zeros(k-1)
    # print("step 1")
    for i in range(1, k):
        theta_i = vertex_vector[i, :]
        theta_0i = 0.5*(theta_0+theta_i)
        chi_sq_0i = func(theta_0i)
        # print("ii: {:}     {:}".format(i, chi_sq_0i))
        m_chi_sq_0i[i-1] = chi_sq_0i
        m_q[i-1, :] = theta_i-theta_0

    # print("step 2")
    for i in range(1, k):
        chi_sq_i = vertex_chi_sq[i]
        theta_i = vertex_vector[i, :]
        chi_sq_0i = m_chi_sq_0i[i-1]

        a_i = 4.*chi_sq_0i - chi_sq_i - 3.*chi_sq_0
        v_a[i-1] = a_i

        b_ii = 2.*(chi_sq_i + chi_sq_0 - 2.*chi_sq_0i)
        m_b[i-1, i-1] = b_ii

        for j in range(i+1, k):
            chi_sq_0j = m_chi_sq_0i[j-1]
            theta_j = vertex_vector[j, :]
            theta_ij = 0.5*(theta_i+theta_j)
            chi_sq_ij = func(theta_ij)
            # print("ij: {:} {:}    {:}".format(i, j, chi_sq_ij))
            b_ij = 2.*(chi_sq_ij + chi_sq_0 - chi_sq_0i - chi_sq_0j)
            m_b[i-1, j-1] = b_ij
            m_b[j-1, i-1] = b_ij
    # print("step 3")
    m_ib = numpy.linalg.inv(m_b)
    m_qib = numpy.matmul(m_q, m_ib)
    v_qiba = numpy.matmul(m_qib, v_a)
    # theta_min = theta_0 - v_qiba
    m_qibqt = numpy.matmul(m_qib, m_q.transpose())
    m_error = 2.*chi_sq_0*m_qibqt

    # print("\nm_q")
    # print(m_q)
    # print("\nm_b")
    # print(m_b)
    # print("\nm_ib")
    # print(m_ib)
    # print("\nv_a")
    # print(v_a)
    # print("\ntheta_min: ", theta_min)
    # print("\ntheta_0: ", theta_0)

    # print("\nm_error: ", m_error)
    # print(50*"*")
    return m_error, numpy.abs(v_qiba)



from .cl_3_data import DataN

def take_items_by_class(global_obj, l_class) -> list:
    l_res = []
    for item in global_obj.items:
        if isinstance(item, l_class):
            l_res.append(item)
        elif isinstance(item, DataN):
            l_res_data = take_items_by_class(item, l_class)
            l_res.extend(l_res_data)
    return l_res


"""RefineLs and RefineLsL classes."""
from typing import NoReturn

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class RefineLs(ItemN):
    """Goodness of fit parameters.

    Data items in the REFINE_LS category record details about the
    structure-refinement parameters.

    By default the UB matrix are given in ccsl notation

    Attributes
    ----------
        - goodness_of_fit_all, number_reflns (mandatory)
        - number_parameters, number_restraints, number_constraints
        - weighting_scheme

    Accesible Values
    ----------------
        - weighting_scheme is "sigma" or "unit" or "calc"
    """

    ATTR_MANDATORY_NAMES = ("goodness_of_fit_all", "number_reflns")
    ATTR_MANDATORY_TYPES = (float, int)
    ATTR_MANDATORY_CIF = ("goodness_of_fit_all", "number_reflns")

    ATTR_OPTIONAL_NAMES = ("number_restraints", "number_constraints",
                           "weighting_scheme", "number_parameters")
    ATTR_OPTIONAL_TYPES = (int, int, str, int)
    ATTR_OPTIONAL_CIF = ("number_restraints", "number_constraints",
                         "weighting_scheme", "number_parameters")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {"goodness_of_fit_all": "{:.3f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {"weighting_scheme": ["sigma", "unit", "calc"]}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "refine_ls"

    def __init__(self, **kwargs) -> NoReturn:
        super(RefineLs, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"number_reflns": 0, "goodness_of_fit_all": 0.,
                 "number_parameters": 0, "number_restraints": 0,
                 "number_constraints": 0}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def report(self):
        l_res = []
        if self.is_attribute("goodness_of_fit_all"):
            l_res.append(f"|Goodness of fit | {self.goodness_of_fit_all:.2f}|")
        if self.is_attribute("number_reflns"):
            l_res.append(f"|Experimental points | {self.number_reflns:}|")
        if self.is_attribute("number_parameters"):
            l_res.append(f"|Parameters number | {self.number_parameters:}|")
        if self.is_attribute("number_restraints"):
            l_res.append(f"|Restraints number | {self.number_restraints:}|")
        if self.is_attribute("number_constraints"):
            l_res.append(f"|Constraints number | {self.number_constraints:}|")
        return "\n".join(l_res)


class RefineLsL(LoopN):
    """Description of RefineLsL class."""

    ITEM_CLASS = RefineLs
    ATTR_INDEX = None

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(RefineLsL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name


# s_cont = """
#   _refine_ls_goodness_of_fit_all        2.74
#   _refine_ls_weighting_scheme           sigma
#   _refine_ls_number_reflns           1408
#   _refine_ls_number_parameters       272
#   _refine_ls_number_restraints       0
#   _refine_ls_number_constraints      0
#   """

# obj = RefineLs.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj.number_reflns, end="\n\n")



from typing import NoReturn
import numpy

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.A_functions_base.matrix_operations import calc_m1_m2, calc_vt_m_v

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN

from cryspy.C_item_loop_classes.cl_1_cell import Cell
from cryspy.C_item_loop_classes.cl_2_diffrn_orient_matrix import DiffrnOrientMatrix

class ReflnSusceptibility(ItemN):
    """Susceptibility structure factor tensor.
    
    Data items in the REFLN_SUSCEPTIBILITY category record details about the
    susceptibility structure factor tensor and magnetization structure factor 
    tensor calculated for different reflections.

    Mandatory attributes:
        - ub_11, ub_12, ub_13, ub_21, ub_22, ub_23, ub_31, ub_32, ub_33

    Optional attributes:
        - occupancy
        - adp_type
        - u_iso_or_equiv
        - u_equiv_geom_mean
        - b_iso_or_equiv
        - multiplicity
        - wyckoff_symbol
        - cartn_x
        - cartn_y
        - cartn_z

    Internal attributes:
        - scat_length_neutron

    Internal protected attributes:
        - space_group_wyckoff
        - constr_number
    """
    ATTR_MANDATORY_NAMES = ("index_h", "index_k", "index_l")
    ATTR_MANDATORY_TYPES = (int, int, int)
    ATTR_MANDATORY_CIF = ("index_h", "index_k", "index_l")

    ATTR_OPTIONAL_NAMES = (
        "sintlambda", "d_spacing", "chi_11_calc", "chi_12_calc", "chi_13_calc", 
        "chi_21_calc", "chi_22_calc", "chi_23_calc", "chi_31_calc",
        "chi_32_calc", "chi_33_calc", )
    ATTR_OPTIONAL_TYPES = (float, float, complex, complex, complex,
                           complex, complex, complex, complex, complex, 
                           complex, complex, complex, complex, )
    ATTR_OPTIONAL_CIF = (
        "sint/lambda", "d_spacing", "chi_11_calc", "chi_12_calc", "chi_13_calc", 
        "chi_21_calc", "chi_22_calc", "chi_23_calc", "chi_31_calc",
        "chi_32_calc", "chi_33_calc", )

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {"sintlambda": "{:.5f}", "d_spacing": "{:.5f}",
                 "chi_11_calc": "{:.2f}", "chi_12_calc": "{:.2f}",
                 "chi_13_calc": "{:.2f}", "chi_21_calc": "{:.2f}",
                 "chi_22_calc": "{:.2f}", "chi_23_calc": "{:.2f}",
                 "chi_31_calc": "{:.2f}", "chi_32_calc": "{:.2f}",
                 "chi_33_calc": "{:.2f}", }

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "refln_susceptibility"

    def __init__(self, **kwargs) -> NoReturn:
        super(ReflnSusceptibility, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"sintlambda": 0, "d_spacing": 0.}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


class ReflnSusceptibilityL(LoopN):
    """Susceptibility structure factor tensors.
    """
    ITEM_CLASS = ReflnSusceptibility
    ATTR_INDEX = "id"
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(ReflnSusceptibilityL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

    def calc_f_mag(self, diffrn_orient_matrix: DiffrnOrientMatrix):
        e_up = diffrn_orient_matrix.calc_e_up(phi=0, omega=0, chi=0)
        m1_ij = numpy.array([
            self.chi_11_calc, self.chi_12_calc, self.chi_13_calc,
            self.chi_21_calc, self.chi_22_calc, self.chi_23_calc,
            self.chi_31_calc, self.chi_32_calc, self.chi_33_calc], dtype=complex)
        
        m2_ij =numpy.conjugate(numpy.stack([
            m1_ij[0], m1_ij[3], m1_ij[6], m1_ij[1], m1_ij[4], m1_ij[7],
            m1_ij[2], m1_ij[5], m1_ij[8]], axis=0))
        
        m_3_ij = (calc_m1_m2(m2_ij, m1_ij))
        f_mag_sq = calc_vt_m_v(m_3_ij, e_up).real
        f_mag = numpy.sqrt(f_mag_sq)
        return f_mag




"""Module define operations with density file: file.den

Functions
---------
    - read_den_file
    - save_to_den_file
    - recalc_den_file_to_p1

"""
import numpy

from cryspy.A_functions_base.function_2_sym_elems import \
    form_symm_elems_by_b_i_r_ij, calc_numerators_denominator_for_b_i, \
    transform_to_p1


def read_den_file(file_name: str):
    """Read density file.
    
    Arguments
    ---------
        -   file_name is a file name str

    points_abc: [3] int
    cell_parameters: [6] float
    sym_elems: [13, n_symm_elems]: int
    indexes_xyz: [3, n_points] int
    densities: [3, n_points] float
    centrosymmetry: bool
    np_shift: [4, n_symmelems]  int
    """
    with open(file_name, "r") as fid:
        l_content = fid.readlines()

    number_lines = int(l_content[1])

    hh = l_content[number_lines+2].strip().split()
    rad = float(numpy.pi/180.)
    cell_parameters = numpy.array([float(hh[0]), float(hh[1]), float(hh[2]), 
                       float(hh[3])*rad, float(hh[4])*rad, float(hh[5])*rad],
                                  dtype=float)

    [points_a, points_b, points_c, n_el_symm, centr, n_shift] = [
        int(hh) for hh in l_content[number_lines+3][:-1].split()]
    points_abc = numpy.array((points_a, points_b, points_c), dtype=int)
    centrosymmetry = bool(centr)

    l_ind_xyz = []
    l_dens = []
    for line in l_content[2:number_lines+2]:
        hh = line.strip().split()
        ind_x, ind_y, ind_z = int(hh[0]), int(hh[1]), int(hh[2])
        den_f, den_a = 0., 0.
        den = float(hh[3])
        if len(hh) == 6:
            den_f = float(hh[4])
            den_a = float(hh[5])
        elif den >= 0.:
            den_f = den
        else:  # den < 0 
            den_a = den
        l_ind_xyz.append((ind_x, ind_y, ind_z))
        l_dens.append((den, den_f, den_a))
    indexes_xyz = numpy.array(l_ind_xyz, dtype=int).transpose()
    densities = numpy.array(l_dens, dtype=float).transpose()

    r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33 = [], [], [], [], \
        [], [], [], [], []
    b_1, b_2, b_3 = [], [], []
    for line in l_content[number_lines+4:number_lines+4+n_el_symm]:
        hh = line.replace("-", " -").strip().split()
        r_11.append(int(hh[0]))
        r_12.append(int(hh[3]))
        r_13.append(int(hh[6]))
        
        r_21.append(int(hh[1]))
        r_22.append(int(hh[4]))
        r_23.append(int(hh[7]))

        r_31.append(int(hh[2]))
        r_32.append(int(hh[5]))
        r_33.append(int(hh[8]))

        b_1.append(float(hh[9]))
        b_2.append(float(hh[10]))
        b_3.append(float(hh[11]))

    b_i = (b_1, b_2, b_3)
    r_ij = (r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33)

    sym_elems = form_symm_elems_by_b_i_r_ij(b_i, r_ij)

    shift_1, shift_2, shift_3 = [], [], []
    for line in l_content[number_lines+4+n_el_symm:
                          number_lines+4+n_el_symm+n_shift]:
        hh = line.strip().split()
        shift_1.append(float(hh[0]))
        shift_2.append(float(hh[1]))
        shift_3.append(float(hh[2]))

    sh_num_x, sh_num_y, sh_num_z, sh_den = calc_numerators_denominator_for_b_i(
        shift_1, shift_2, shift_3)

    np_shift = numpy.stack((sh_num_x, sh_num_y, sh_num_z, sh_den), axis=0)

    return points_abc, cell_parameters, sym_elems, indexes_xyz, densities, \
        centrosymmetry, np_shift


def save_to_den_file(
        file_name: str, points_abc, cell_parameters, sym_elems, indexes_xyz,
        densities, centrosymmetry, shift):
    """Save to file.
    file_name is str
    points_abc: [3] int
    cell_parameters: [6] float
    sym_elems: [13, n_symm_elems]: int
    indexes_xyz: [3, n_points] int
    densities: [3, n_points] float
    centrosymmetry: bool
    np_shift: [4, n_symmelems]  int
    """

    index_x = indexes_xyz[0, :]
    index_y = indexes_xyz[1, :]
    index_z = indexes_xyz[2, :]

    n_symmetry = sym_elems.shape[1]

    n_shape = densities.shape
    if len(n_shape) == 1:
        n_points = n_shape[0]
        n_size = 1 
    else:
        n_size, n_points = n_shape
    if n_size == 3:
        density = densities[0, :]
        density_f = densities[1, :]
        density_a = densities[2, :]
    elif n_size == 2:
        density_f = densities[0, :]
        density_a = densities[1, :]
        density = density_f-density_a
    else:
        density = densities
        density_f = numpy.zeros(density.shape, dtype=float)
        density_a = numpy.zeros(density.shape, dtype=float)
        cond = density >= 0.
        density_f[cond] = density[cond]
        cond_not = numpy.logical_not(cond)
        density_a[cond_not] = density[cond_not]

    ls_out = []
    ls_out.append("Created by CrysPy")
    ls_out.append("{:}".format(n_points))

    for i_x, i_y, i_z, den, den_f, den_a in zip(
            index_x, index_y, index_z, density, density_f, density_a):

        ls_out.append(
            f"{i_x:4} {i_y:4} {i_z:4} {den:15.7f} {den_f:15.7f} {den_a:15.7f}")

    (a, b, c, al, be, ga) = cell_parameters

    ls_out.append(
        f"{a:10.5f}{b:10.5f}{c:10.5f}{al:10.5f}{be:10.5f}{ga:10.5f}")

    n_shift = shift.shape[1]

    ls_out.append(f"{points_abc[0]:5}{points_abc[1]:5}{points_abc[2]:5}\
{n_symmetry:5}{centrosymmetry:5}{n_shift:5}")

    for r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33, b_num_1,  \
            b_num_2, b_num_3, b_den in zip(
            sym_elems[4, :], sym_elems[5, :], sym_elems[6, :], 
            sym_elems[7, :], sym_elems[8, :], sym_elems[9, :], 
            sym_elems[10, :], sym_elems[11, :], sym_elems[12, :], 
            sym_elems[0, :], sym_elems[1, :], sym_elems[2, :],
            sym_elems[3, :]):
        b_1 = float(b_num_1)/float(b_den)
        b_2 = float(b_num_2)/float(b_den)
        b_3 = float(b_num_3)/float(b_den)
        ls_out.append(
            f"{r_11:4}{r_21:4}{r_31:4}  {r_12:4}{r_22:4}{r_32:4}  \
{r_13:4}{r_23:4}{r_33:4}    {b_1:8.5f}{b_2:8.5f}{b_3:8.5f}")
    for orig_1, orig_2, orig_3 in zip(shift[0, :], shift[1, :], shift[2, :]):
        ls_out.append(f"{orig_1:8.4f}{orig_2:8.4f}{orig_3:8.4f}")

    with open(file_name, "w") as fid:
        fid.write("\n".join(ls_out))

def recalc_den_file_to_p1(den_file_input: str, den_file_output: str):
    points_abc, cell_parameters, sym_elems, indexes_xyz, densities, \
        centrosymmetry, np_shift = read_den_file(den_file_input)

    sym_elems_p1 = numpy.array([[0], [0], [0], [1], [1], [0], [0], [0], [1],
                                [0], [0], [0], [1]], dtype=int)

    np_shift_p1 = numpy.array([[0], [0], [0], [1]], dtype=int)
    indexes_xyz_p1, densities_p1 = transform_to_p1(
        points_abc, sym_elems, np_shift, centrosymmetry, indexes_xyz, densities)

    save_to_den_file(
        den_file_output, points_abc, cell_parameters, sym_elems_p1,
        indexes_xyz_p1, densities_p1, False, np_shift_p1)



# -*- coding: utf-8 -*-
"""Console mode of cryspy library."""
import os
import sys

from cryspy.H_functions_global.function_1_cryspy_objects import \
    file_to_globaln, RhoChi, MEM

if __name__ == "__main__":
    print("Console mode of cryspy library")
    l_arg = sys.argv
    if len(l_arg) > 1:
        f_name = l_arg[1]
        if not os.path.isfile(f_name):
            f_name = "main.rcif"
    if not os.path.isfile(f_name):
        print("File is not found")
    else:
        obj_global = file_to_globaln(f_name)
        if isinstance(obj_global, RhoChi):
            print("RhoChi object is detected")
            print("Refienement is running")
            res = obj_global.refine()
            print("Result of refinement: \n ", res)
            obj_global.save_to_file(f_name)
            print("The file is rewritten")
        elif isinstance(obj_global, MEM):
            print("MEM object is detected")
            print("The action algorithm is not written.")
        else:
            print("The found file is not defined as any predescribed class.")



"""
Describing of parent class LoopN.

It should be used ass parent to describe new loop classes.
"""
import os
import os.path
import copy
from typing import NoReturn, Union
import numpy
from pycifstar import Data, to_data

from cryspy.A_functions_base.function_1_markdown import md_to_html
from cryspy.A_functions_base.function_1_strings import find_prefix, \
    string_to_value_error_mark
from cryspy.A_functions_base.function_1_objects import \
    get_functions_of_objet, get_table_html_for_variables

from cryspy.B_parent_classes.cl_1_item import ItemN


class LoopN(object):
    """Loop data.

    It is internal class of cryspy library.
    You should use it only to create your own classes.
    """

    def __repr__(self):
        """
        Magic method print() is redefined.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        ls_out = [f"# Object '{self.get_name():}':\n"]
        ls_out.append(str(self))

        method = self.methods_html()
        if method != "":
            ls_out.append(f"\n# Methods:\n{method:}\n")                
        return "\n".join(ls_out)

    def _repr_html_(self):
        """Representation in HTML format."""
        ls_html = [f"<h3>Object '{self.get_name():}'</h3>"]
        ls_html.append(self.attributes_to_html())

        ls_html.append(get_table_html_for_variables(self))
        
        report = self.report_html()
        if report != "":
            ls_html.append(f"<h3>Description </h3> {report:}")

        ls_html.append(f"<h3>Methods</h3>")
        method = self.methods_html()
        if method != "":
            ls_html.append(f"<b>Methods: </b> {method:}")

        return " ".join(ls_html)

    def methods_html(self):
        ls_html = [f".{func_name}" for func_name in
                   get_functions_of_objet(self)]
        return ", ".join(ls_html)+"."
    
    def __str__(self):
        """
        Magic method str() is redefined.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        return self.to_cif()

    def __getattr__(self, name):
        """
        Magic method __getattr__ is slightly changed for special attributes.

        Parameters
        ----------
        name : TYPE
            DESCRIPTION.

        Raises
        ------
        AttributeError
            DESCRIPTION.

        Returns
        -------
        res : TYPE
            DESCRIPTION.

        """
        item_class = self.ITEM_CLASS
        if item_class is ItemN:
            item_class = self.items[0]
        if name.startswith("numpy_"):
            name_sh = name[6:]
            if name_sh in (item_class.ATTR_NAMES + item_class.ATTR_SIGMA +
                           item_class.ATTR_CONSTR_FLAG +
                           item_class.ATTR_REF_FLAG + item_class.ATTR_CONSTR_MARK):
                l_val = [getattr(item, name_sh) for item in self.items]
                if name_sh in item_class.ATTR_NAMES:
                    type_array = item_class.ATTR_TYPES[
                        item_class.ATTR_NAMES.index(name_sh)]
                elif name_sh in (item_class.ATTR_CONSTR_FLAG +
                                 item_class.ATTR_REF_FLAG):
                    type_array = bool
                elif name_sh in item_class.ATTR_CONSTR_MARK:
                    type_array = str
                else:
                    type_array = float
                res = numpy.array(l_val, dtype=type_array)
                self.__dict__[name] = res
                return res
            elif name_sh in (item_class.ATTR_INT_NAMES +
                             item_class.ATTR_INT_PROTECTED_NAMES):
                l_val = [getattr(item, name_sh) for item in self.items]
                type_array = type(l_val[0])
                res = numpy.array(l_val, dtype=type_array)
                self.__dict__[name] = res
                return res
        elif name in (item_class.ATTR_NAMES + item_class.ATTR_SIGMA +
                      item_class.ATTR_CONSTR_FLAG + item_class.ATTR_REF_FLAG +
                      item_class.ATTR_CONSTR_MARK):
            res = [getattr(item, name) for item in self.items]
            return res
        elif name in (item_class.ATTR_INT_NAMES +
                      item_class.ATTR_INT_PROTECTED_NAMES):
            res = [getattr(item, name) for item in self.items]
            return res
        raise AttributeError(
            f"'{type(self).__name__:}' object has no attribute '{name:}'")

    def __setattr__(self, name, value) -> NoReturn:
        """
        Rules to set attribute.

        Parameters
        ----------
        name : TYPE
            DESCRIPTION.
        value : TYPE
            DESCRIPTION.

        Returns
        -------
        NoReturn
            DESCRIPTION.

        """
        flag_direct = True
        if name == "items":
            flag = all([isinstance(val, self.ITEM_CLASS) for val in value])
            if not(flag):
                flag_direct = False
                val_new = [val for val in value if isinstance(val,
                                                              self.ITEM_CLASS)]
        elif name == "loop_name":
            flag_direct = False
            val_new = str(value).strip()

        if flag_direct:
            self.__dict__[name] = value
        else:
            self.__dict__[name] = val_new

    def is_attribute(self, name: str):
        """Give True if all attributes are defined."""
        flag = all([item.is_attribute(name) for item in self.items])
        return flag

    def numpy_to_items(self) -> NoReturn:
        """
        Transform all internal numpy arrays to elements of items.

        Returns
        -------
        NoReturn
            DESCRIPTION.

        """
        loop_dict = self.__dict__
        item_class = self.ITEM_CLASS
        keys = loop_dict.keys()
        del_name = []
        for name in keys:
            if name.startswith("numpy_"):
                name_sh = name[6:]
                if name_sh in (item_class.ATTR_NAMES + item_class.ATTR_SIGMA +
                               item_class.ATTR_CONSTR_FLAG +
                               item_class.ATTR_REF_FLAG + 
                               item_class.ATTR_CONSTR_MARK):
                    numpy_val = self.__dict__[name]
                    if len(self.items) == 0:
                        items = [item_class() for val in numpy_val]
                        self.items = items
                    for item, val in zip(self.items, numpy_val):
                        setattr(item, name_sh, val)
                    del_name.append(name)
        for name in del_name:
            del self.__dict__[name]

    def form_object(self):
        pass

    @classmethod
    def define_by_items(cls, l_item: list, loop_name: str = ""):
        """Define LoopN by defined items.
        
        (it's for the future)
        """
        if len(l_item) == 0:
            raise ValueError("Number of items is zero")
        if not(isinstance(l_item[0], ItemN)):
            raise ValueError("type of items is not ItemN")
        item_class = type(l_item[0])
        flag = all([type(item) is item_class for item in l_item])
        if not(flag):
            raise ValueError("There are different types of given items")
        obj = cls()
        if cls is LoopN:
            obj.__dict__["ITEM_CLASS"] = item_class
            obj.__dict__["ATTR_INDEX"] = None
            obj.__dict__["items"] = l_item
            obj.__dict__["loop_name"] = loop_name
        elif obj.ITEM_CLASS is item_class:
            obj.items = l_item
        else:
            raise ValueError("Type of items does not correspond to loop class")            
        return obj

            
    @classmethod
    def from_cif(cls, string: str):
        """
        Create object of Cell class from string.

        Parameters
        ----------
        cls : TYPE
            DESCRIPTION.
        string : str
            DESCRIPTION.

        Returns
        -------
        obj : TYPE
            DESCRIPTION.

        """
        cif_data = Data()
        cif_data.take_from_string(string)
        obj = None
        if cls is LoopN:
            loop = cif_data.loops[0]
            l_name = loop.names
            prefix = find_prefix(*tuple(l_name))
            l_attr = [name[len(prefix):].strip("_").strip(".") for name in l_name]
            l_cif_attr = [_.lower() for _ in l_attr]
            if "" in l_attr:
                l_attr[l_attr.index("")] = "label"
            prefix = prefix.strip("_").strip("_")
            obj = cls()
            obj.__dict__["ITEM_CLASS"] = ItemN
            obj.__dict__["ATTR_INDEX"] = None
            item_class = ItemN
            obj.__dict__["items"] = []
            obj.__dict__["loop_name"] = loop.name
        else:
            item_class = cls.ITEM_CLASS
            l_attr = item_class.ATTR_NAMES
            l_cif_attr = item_class.ATTR_CIF

            l_cif_attr = [_.lower() for _ in l_cif_attr]
            prefix = item_class.PREFIX

        for cif_loop in cif_data.loops:
            # cond = ("_" + prefix.lower()) == cif_loop.prefix.lower()
            # print("cif_loop.prefix: ", cif_loop.prefix)
            # print(("_" + prefix.lower()) == cif_loop.prefix.lower(),
            #       cif_loop.prefix.lower().startswith("_" + prefix.lower()))
            # if ("_" + prefix.lower()) == cif_loop.prefix.lower():
            if (cif_loop.prefix.lower().startswith("_" + prefix.lower())):
                prefix = f"_{prefix:}"
                # prefix = cif_loop.prefix
                l_name = cif_loop.names
                loop_name = cif_loop.name
                l_name_short = [_[(len(prefix) + 1):] for _ in l_name]

                if obj is None:
                    obj = cls(loop_name=loop_name)
                _i = 0
                for _name, _name_short in zip(l_name, l_name_short):
                    if _name_short.lower() in l_cif_attr:
                        _name_short_obj = l_attr[l_cif_attr.index(_name_short)]
                    else:
                        _name_short_obj = _name_short.lower()

                    if _name_short_obj.strip("_").strip(".") in l_attr:
                        if _i == 0:
                            items = []
                            for _val in cif_loop[_name]:
                                if item_class is ItemN:
                                    item = item_class()
                                    item._base_attributes_itemn(
                                        prefix.strip("_").strip("."),
                                        tuple(l_cif_attr), tuple(l_attr))
                                else:
                                    item = item_class()

                                if _name_short_obj in item.ATTR_REF:
                                    value, error, mark = string_to_value_error_mark(_val)
                                    
                                    setattr(item, _name_short_obj, value)
                                    if error is not None:
                                        setattr(item, f"{_name_short_obj:}_sigma", error)
                                        setattr(item, f"{_name_short_obj:}_refinement", True)
                                        setattr(item, f"{_name_short_obj:}_mark", mark)
                                    else:
                                        setattr(
                                            item,
                                            f"{_name_short_obj:}_refinement",
                                            False)
                                else:
                                    setattr(item, _name_short_obj, _val)
                                items.append(item)
                        else:
                            for _val, item in zip(cif_loop[_name], items):
                                if _name_short_obj in item.ATTR_REF:
                                    value, error, mark = string_to_value_error_mark(_val)
                                    setattr(item, _name_short_obj, value)
                                    if error is not None:
                                        setattr(item, f"{_name_short_obj:}_sigma", error)
                                        setattr(item,f"{_name_short_obj:}_refinement", True)
                                        setattr(item,f"{_name_short_obj:}_mark", mark)
                                    else:
                                        setattr(item, f"{_name_short_obj:}_refinement", False)
                                else:
                                    setattr(item, _name_short_obj, _val)
                        _i += 1
                    else:
                        pass

                obj.items = items
                break

        return obj

    def to_cif(self, separator: str = "_") -> str:
        """
        Give data in CIF/STAR format.

        Parameters
        ----------
        separator : str, optional
            DESCRIPTION. The default is "_".

        Returns
        -------
        str
            DESCRIPTION.

        """
        ls_out = []
        item_class = self.ITEM_CLASS
        if self.loop_name is None:
            ls_out.append("loop_")
        else:
            ls_out.append(f"loop_{self.loop_name:}")

        if len(self.items) == 0:
            if item_class == ItemN:
                return ""
            prefix = item_class.PREFIX
            for name_cif in item_class.ATTR_CIF:
                ls_out.append(f"_{prefix:}{separator:}{name_cif:}")
            ls_out.append(" ".join(len(item_class.ATTR_CIF)*["."]))
            return "\n".join(ls_out)

        item_0 = self.items[0]
        prefix = item_0.PREFIX
        ls_out_2 = []
        for name, name_cif in zip(item_0.ATTR_NAMES, item_0.ATTR_CIF):
            flag_value = item_0.is_attribute(name)
            if flag_value:
                if name_cif == "":
                    ls_out.append(f"_{prefix:}")
                else:
                    ls_out.append(f"_{prefix:}{separator:}{name_cif:}")
                list_value = [getattr(item, f"{name}_as_string")
                              for item in self.items]
                if len(ls_out_2) == 0:
                    n_max = max(map(len, [str(h) for h in list_value])) + 2
                    for s_val in list_value:
                        if ((len(s_val.split(" ")) > 1) |
                                s_val.startswith("_")):
                            ls_out_2.append([("\""+s_val+"\"").rjust(n_max)])
                        else:
                            ls_out_2.append([s_val.rjust(n_max)])
                else:
                    n_max = max(map(len, [str(h) for h in list_value])) + 2
                    for list_out, s_val in zip(ls_out_2, list_value):
                        if ((len(s_val.split(" ")) > 1) | 
                                s_val.startswith("_")):
                            list_out.append(("\""+s_val+"\"").rjust(n_max))
                        else:
                            list_out.append(s_val.rjust(n_max))

        for list_out in ls_out_2:
            s_line = " ".join(list_out)
            ls_out.append(s_line)
        return "\n".join(ls_out)

    def attributes_to_html(self) -> str:
        """Representation of defined parameters in HTML format.
        """
        ls_html = ["<table>"]
        item_class = self.ITEM_CLASS
        if item_class is ItemN:
            if len(self.items) != 0:
                item_n = self.items[0]
                attribute_names = item_n.ATTR_NAMES
            else:
                return ""
        else:
            attribute_names = item_class.ATTR_NAMES
        flags = [False for name in attribute_names]
        for item in self.items:
            for ind_name, name in enumerate(attribute_names):
                flag_value = item.is_attribute(name)
                if flag_value:
                    flags[ind_name] = True

        ls_html.append(
            "<tr><th> </th>"+"".join([f"<th>.{name:}</th>" for name, flag in zip(
                attribute_names, flags)  if flag])+"</tr>")

        for ind_item, item in enumerate(self.items):
            ls_line = []
            for name, flag in zip(attribute_names, flags):
                if flag:
                    flag_value = item.is_attribute(name)
                    if flag_value:
                        value = getattr(item, f"{name}_as_string")
                        ls_line.append(f"<td>{value:}</td>")
                    else:
                        ls_line.append("<td>.</td>")
            ls_html.append(f"<tr><th>[{ind_item}]</th>"+"".join(ls_line)+"</tr>")
            if ind_item > 10:
                ls_line = ["<td>...</td>" for flag in flags if flag ]
                ls_html.append(f"<tr><th>[..., {len(self.items):}]</th>"+"".join(ls_line)+
                               "</tr>")
                break

        ls_html.append("</table>")
        return " ".join(ls_html)

    def save_to_csv(self, file_name: str = "output.csv", delimiter: str = ";")\
            -> NoReturn:
        """
        Save loop into csv format.

        Parameters
        ----------
        file_name : str, optional
            Output file name. The default is "output.csv".
        delimiter : str, optional
            Delimeter. The default is ";".

        Returns
        -------
        NoReturn

        """
        item_class = self.ITEM_CLASS
        l_attr_print = item_class.ATTR_NAMES
        l_cif_attr = item_class.ATTR_NAMES_CIF
        """
        l_flag = [all([_item.is_defined_attribute(_attr) for _item in self.item]) for _attr in l_attr_print]
        l_attr_print = [_ for _, _flag in zip(l_attr_print, l_flag) if _flag]
        l_cif_attr = [_ for _, _flag in zip(l_cif_attr, l_flag) if _flag]
        ls_out = [(f"{delimiter:}").join([f"{_attr:}" for _attr in l_cif_attr])]
        for _item in self.item:
            ls_out.append(_item.print_attribute(l_attr_print, delimiter=delimiter))
        s_cont = "\n".join(ls_out)
        with open(file_name, "w") as fid:
            fid.write(s_cont)
        """

    def __getitem__(self, name):
        """
        Get item by index or predefined index.

        Parameters
        ----------
        name : TYPE
            DESCRIPTION.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        attr_index = self.ATTR_INDEX
        if attr_index is None:
            if isinstance(name, int):
                return self.items[name]
            else:
                return None
        else:
            for item in self.items:
                if name == getattr(item, attr_index):
                    return item
            if isinstance(name, int):
                return self.items[name]

    def get_name(self) -> str:
        """Get name."""
        if self.ITEM_CLASS is ItemN:
            if len(self.items) > 0:
                name = self.items[0].PREFIX
            else:
                return ""
        else:
            name = self.ITEM_CLASS.PREFIX
        loop_name = self.loop_name
        if ((loop_name is None) | (loop_name == "")):
            name = f"{name:}"
        else:
            name = f"{name:}_{loop_name:}"
        return name.lower()

    def get_variable_names(self) -> list:
        """
        Get names of variable as a list.

        (((#prefix, #NAME), (#attribute, #index))

        Returns
        -------
        list
            List of names of variable.

        """
        item_class = self.ITEM_CLASS
        if item_class is ItemN:
            if len(self.items) != 0:
                prefix = self.items[0].PREFIX
            else:
                return []
        else:
            prefix = item_class.PREFIX
        loop_name = self.loop_name
        if isinstance(loop_name, str):
            loop_name = loop_name.lower()
        l_var = []
        for ind, item in enumerate(self.items):
            l_var.extend([((prefix, loop_name), (name[1][0], ind))
                          for name in item.get_variable_names()])
        return l_var

    def is_variables(self) -> bool:
        """Define is there variables or not."""
        flag = False
        for item in self.items:
            if item.is_variables():
                flag = True
                break
        return flag

    def get_variable_by_name(self, name: tuple) -> Union[float, int, str]:
        """
        Get variable given by name.

        Parameters
        ----------
        name : tuple
            (((#prefix, #loop_name), (#attribute, #index_item))

        Returns
        -------
        Union[float, int, str]
            DESCRIPTION.

        """
        item_class = self.ITEM_CLASS
        if item_class is ItemN:
            if len(self.items) != 0:
                prefix = self.items[0].PREFIX
            else:
                return
        else:
            prefix = item_class.PREFIX
        prefix_t = name[0]
        if isinstance(prefix_t[1], str):
            prefix_t_2 = prefix_t[1].lower()
        else:
            prefix_t_2 = prefix_t[1]

        loop_name = self.loop_name
        if isinstance(loop_name, str):
            loop_name = loop_name.lower()

        if (prefix_t[0], prefix_t_2) != (prefix, loop_name):
            return None

        if len(name) == 1:
            return self
        attr_t = name[1]
        attr_name = attr_t[0]
        if len(attr_t) == 1:
            return getattr(self.items, attr_name)
        item_index = attr_t[1]
        if attr_name is None:
            return self.items[item_index]
        return getattr(self.items[item_index], attr_name)

    def set_variable_by_name(self, name: tuple, value) -> NoReturn:
        """
        Set value to variable given by name.

        Parameters
        ----------
        name : tuple
            DESCRIPTION.
        value : TYPE
            DESCRIPTION.

        Returns
        -------
        NoReturn
            DESCRIPTION.

        """
        item_class = self.ITEM_CLASS
        if item_class is ItemN:
            if len(self.items) != 0:
                prefix = self.items[0].PREFIX
            else:
                return
        else:
            prefix = item_class.PREFIX
        
        prefix_t, attr_t = name
        if prefix_t != (prefix, self.loop_name):
            return
        attr_name, item_index = attr_t
        setattr(self.items[item_index], attr_name, value)

    def is_defined(self) -> bool:
        """
        If all mandatory attributes is defined.

        Returns
        -------
        bool
            DESCRIPTION.

        """
        flag = True
        for item in self.items:
            if not(item.is_defined()):
                flag = False
                break
        return flag

    @classmethod
    def from_cif_file(cls, f_name: str):
        """Read from cif file."""
        if not(os.path.isfile(f_name)):
            raise UserWarning(f"File {f_name:} is not found.")
            return None
        str_from_cif = str(to_data(f_name))
        obj = cls.from_cif(str_from_cif)
        obj.file_input = f_name
        return obj

    def copy_from(self, obj):
        """Copy attributes from obj to self."""
        
        if type(obj) is not type(self):
            return
        self.items.clear()
        self.items = [type(item).from_cif(str(item))  for item in obj.items]
        
        

    def report(self):
        return ""

    def report_html(self):
        return md_to_html(self.report())

    def plots(self):
        return []

    def fix_variables(self):
        """Fix variables."""
        for item in self.items:
            item.fix_variables()

    def refine_all_variables(self):
        for item in self.items:
            item.refine_all_variables()

    def set_variable(self, name: str, index=None):
        """Set refinement for variable given by name.
        
        Index parameters is used only for objects given as a matrix.
        """
        name_sh = name.strip(".").lower()
        item_class = self.ITEM_CLASS
        if item_class is ItemN:
            if len(self.items) != 0:
                att_ref = self.items[0].ATTR_REF
            else:
                return
        else:
            att_ref = item_class.ATTR_REF

        if name_sh in att_ref:
            if index is None:
                for item in self.items:
                    item.set_variable(name_sh, index=index)
            else:
                item = self[index]
                item.set_variable(name_sh, index=None)
    
    def get_dictionary(self):
        res = {}
        return res
    
    def take_parameters_from_dictionary(self, ddict_diffrn, l_parameter_name: list=None, l_sigma: list=None):
        """
        """
        pass
        return None

def get_prefix_of_loop(item: LoopN):
    if not(isinstance(item, LoopN)):
        return ""
    if item.ITEM_CLASS is ItemN:
        if len(item.items) > 0:
            prefix = item.items[0].PREFIX
        else:
            return ""
    else:
        prefix = item.ITEM_CLASS.PREFIX
    return prefix


from typing import NoReturn
import numpy

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class Texture(ItemN):
    """
    Texture description by Modified March’s function.

    Mandatory Attributes
    --------------------
        - g_1, g_2, h_ax, k_ax, l_ax

    """
    ATTR_MANDATORY_NAMES = ("g_1", "g_2", "h_ax", "k_ax", "l_ax")
    ATTR_MANDATORY_TYPES = (float, float, float, float, float)
    ATTR_MANDATORY_CIF = ("g_1", "g_2", "h_ax", "k_ax", "l_ax")

    ATTR_OPTIONAL_NAMES = ("label", )
    ATTR_OPTIONAL_TYPES = (str, )
    ATTR_OPTIONAL_CIF = ("label", )

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("g_1", "g_2", "h_ax", "k_ax", "l_ax")
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {'g_1': "{:.5f}", 'g_2': "{:.5f}", 'h_ax': "{:.3f}",
                 'k_ax': "{:.3f}", 'l_ax': "{:.3f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "texture"

    def __init__(self, **kwargs) -> NoReturn:
        super(Texture, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


class TextureL(LoopN):
    """
    Description of Texture in loop.

    """
    ITEM_CLASS = Texture
    ATTR_INDEX = None
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(TextureL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

    def get_dictionary(self):
        res = {}
        res["texture_name"] = numpy.array(self.label, dtype=str)
        res["texture_g1"] = numpy.array(self.g_1, dtype=float)
        res["texture_g2"] = numpy.array(self.g_2, dtype=float)
        res["texture_axis"] = numpy.array(
            [self.h_ax, self.k_ax, self.l_ax], dtype=float)
        res["flags_texture_g1"] = numpy.array(self.g_1_refinement, dtype=bool)
        res["flags_texture_g2"] = numpy.array(self.g_2_refinement, dtype=bool)
        res["flags_texture_axis"] = numpy.array(
            [self.h_ax_refinement,
                self.k_ax_refinement,
                self.l_ax_refinement], dtype=bool)
        return res


import numpy
from typing import NoReturn

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class TOFBackground(ItemN):
    """Background description for time-of-flight experiment.

    Attributes
    ----------
        - time_max, coeff1 (mandatory)
        - coeff2, coeff3, coeff4, coeff5, coeff6, coeff7, coeff8,
          coeff9, coeff10, coeff11, coeff12, coeff13, coeff14, coeff15,
          coeff16, coeff17, coeff18, id (optional)
    """
    ATTR_MANDATORY_NAMES = ("time_max", "coeff1")
    ATTR_MANDATORY_TYPES = (float, float)
    ATTR_MANDATORY_CIF = ("time_max", "coeff1")

    ATTR_OPTIONAL_NAMES = (
        "coeff2", "coeff3", "coeff4", "coeff5", "coeff6", "coeff7", "coeff8",
        "coeff9", "coeff10", "coeff11", "coeff12", "coeff13", "coeff14",
        "coeff15", "coeff16", "coeff17", "coeff18", "id")

    ATTR_OPTIONAL_TYPES = (
        float, float, float, float, float, float, float, float, float, float,
        float, float, float, float, float, float, float, str)
    ATTR_OPTIONAL_CIF = (
        "coeff2", "coeff3", "coeff4", "coeff5", "coeff6", "coeff7","coeff8",
        "coeff9", "coeff10", "coeff11", "coeff12", "coeff13", "coeff14",
        "coeff15", "coeff16", "coeff17", "coeff18", "id")


    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = (
        "coeff1", "coeff2", "coeff3", "coeff4", "coeff5", "coeff6", "coeff7",
        "coeff8", "coeff9", "coeff10", "coeff11", "coeff12", "coeff13",
        "coeff14", "coeff15", "coeff16", "coeff17", "coeff18")
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {'time_max': "{:.2f}",
                 'coeff1': "{:.5f}", 'coeff2': "{:.5f}", 'coeff3': "{:.5f}",
                 'coeff4': "{:.5f}", 'coeff5': "{:.5f}", 'coeff6': "{:.5f}",
                 'coeff7': "{:.5f}", 'coeff8': "{:.5f}", 'coeff9': "{:.5f}",
                 'coeff10': "{:.5f}", 'coeff11': "{:.5f}", 'coeff12': "{:.5f}",
                 'coeff13': "{:.5f}", 'coeff14': "{:.5f}", 'coeff15': "{:.5f}",
                 'coeff16': "{:.5f}", 'coeff17': "{:.5f}", 'coeff18': "{:.5f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "tof_background"

    def __init__(self, **kwargs) -> NoReturn:
        super(TOFBackground, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"time_max": 0.01}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def calc_background(self, time):
        """
        y = \sum_{i=1}^{18} coeff_i cos (pi i * (time/time_max))

        Parameters
        ----------
        time : TYPE
            DESCRIPTION.

        Returns
        -------
        y: background points

        """
        time_rel = (time/self.time_max)*numpy.pi
        np_cos = numpy.cos(0.*time_rel)
        res = self.coeff1 * np_cos
        if self.is_attribute("coeff2"):
            np_cos = numpy.cos(1.*time_rel)
            res += self.coeff2 * np_cos
        if self.is_attribute("coeff3"):
            np_cos = numpy.cos(2.*time_rel)
            res += self.coeff3 * np_cos
        if self.is_attribute("coeff4"):
            np_cos = numpy.cos(3.*time_rel)
            res += self.coeff4 * np_cos
        if self.is_attribute("coeff5"):
            np_cos = numpy.cos(4.*time_rel)
            res += self.coeff5 * np_cos
        if self.is_attribute("coeff6"):
            np_cos = numpy.cos(5.*time_rel)
            res += self.coeff6 * np_cos
        if self.is_attribute("coeff7"):
            np_cos = numpy.cos(6.*time_rel)
            res += self.coeff7 * np_cos
        if self.is_attribute("coeff8"):
            np_cos = numpy.cos(7.*time_rel)
            res += self.coeff8 * np_cos
        if self.is_attribute("coeff9"):
            np_cos = numpy.cos(8.*time_rel)
            res += self.coeff9 * np_cos
        if self.is_attribute("coeff10"):
            np_cos = numpy.cos(9.*time_rel)
            res += self.coeff10 * np_cos
        if self.is_attribute("coeff11"):
            np_cos = numpy.cos(10.*time_rel)
            res += self.coeff11 * np_cos
        if self.is_attribute("coeff12"):
            np_cos = numpy.cos(11.*time_rel)
            res += self.coeff12 * np_cos
        if self.is_attribute("coeff13"):
            np_cos = numpy.cos(12.*time_rel)
            res += self.coeff13 * np_cos
        if self.is_attribute("coeff14"):
            np_cos = numpy.cos(13.*time_rel)
            res += self.coeff14 * np_cos
        if self.is_attribute("coeff15"):
            np_cos = numpy.cos(14.*time_rel)
            res += self.coeff15 * np_cos
        if self.is_attribute("coeff16"):
            np_cos = numpy.cos(15.*time_rel)
            res += self.coeff16 * np_cos
        if self.is_attribute("coeff17"):
            np_cos = numpy.cos(16.*time_rel)
            res += self.coeff17 * np_cos
        if self.is_attribute("coeff18"):
            np_cos = numpy.cos(17.*time_rel)
            res += self.coeff18 * np_cos
        return res

    def get_coefficients(self):
        l_coeff = []
        last_number = 1
        for numb in range(1, 19):
            if self.is_attribute(f"coeff{numb:}"):
                coeff = getattr(self, f"coeff{numb:}")
                try:
                    coeff = float(coeff)
                    last_number = numb
                except TypeError:
                    coeff = 0
            else:
                coeff = 0.
            l_coeff.append(coeff)
        coefficients = numpy.array(l_coeff[:last_number], dtype=float)
        return coefficients

    def get_flags_coefficients(self):
        l_flag_coeff = []
        last_number = 1
        for numb in range(1, 19):
            if self.is_attribute(f"coeff{numb:}"):
                flag_coeff = getattr(self, f"coeff{numb:}_refinement")
                last_number = numb
            else:
                flag_coeff = False
            l_flag_coeff.append(flag_coeff)
        flags_coefficients = numpy.array(l_flag_coeff[:last_number], dtype=bool)
        return flags_coefficients

    def get_dictionary(self):
        res = {}
        res["background_time_max"] = getattr(self, "time_max")
        res["background_coefficients"] = self.get_coefficients()
        res["flags_background_coefficients"] = self.get_flags_coefficients()
        return res

class TOFBackgroundL(LoopN):
    """Bacground description for time-of-flight experiment.

    """
    ITEM_CLASS = TOFBackground
    ATTR_INDEX = "id"
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(TOFBackgroundL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name


# s_cont = """
# _tof_background_time_max 30000
# _tof_background_coeff1 24832.850
# _tof_background_coeff2 6139.244
# _tof_background_coeff3 8063.472
# _tof_background_coeff4 3125.050
# _tof_background_coeff5 2566.956
# _tof_background_coeff6 311.077
# _tof_background_coeff7 837.348
# _tof_background_coeff8 -103.742
# _tof_background_coeff9 -11.806
# """

# obj = TOFBackground.from_cif(s_cont)
# print(obj, end="\n\n")
# time = numpy.linspace(1, 30000, 10)
# y = obj.calc_background(time)
# print("time bbackground")
# for h1, h2 in zip(time, y):
#     print(f"{h1:10.2f} {h2:10.2f}")



"""Mactices functions.

Functions
---------
    - calc_chi_sq
    - transform_string_to_r_b
    - transform_string_to_digits
    - transform_fraction_with_label_to_string
    - transform_digits_to_string
    - transform_r_b_to_string
    - calc_mRmCmRT
    - calc_rotation_matrix_ij_by_euler_angles
    - calc_euler_angles_by_rotation_matrix_ij
    - calc_determinant_matrix_ij
    - calc_inverse_matrix_ij
    - calc_rotation_matrix_ij_around_axis
    - calc_product_matrices
    - calc_product_matrix_vector
    - calc_vector_angle
    - calc_vector_product
    - scalar_product
    - calc_rotation_matrix_by_two_vectors
    - tri_linear_interpolation
    - ortogonalize_matrix
    - calc_moment_2d_by_susceptibility
    - calc_phase_3d
"""
import numpy
from fractions import Fraction
from typing import Tuple


def calc_chi_sq(y_exp, y_error, y_model, der_y=None):
    """Calc chi_sq."""
    y_diff = (y_exp - y_model)/y_error
    chi_sq = (numpy.square(numpy.abs(y_diff))).sum()
    if der_y is not None:
        der_chi_sq = (-2.*(y_diff/y_error)[:, numpy.newaxis]*der_y).sum(axis=0)
    else:
        der_chi_sq = None
    return chi_sq, der_chi_sq


def transform_string_to_r_b(name: str, labels=("x", "y", "z")) -> Tuple:
    """Transform string to rotation part and offset.

    Example
    -------
    x,y,-z   ->   0.0 1 0 0  0.0 0 1 0  0.0 0 0 -1
    """
    l_name = "".join(name.strip().split()).lstrip("(").rstrip(")").split(",")
    rij, bi = [], []
    for _name in l_name:
        coefficients, offset = transform_string_to_digits(_name, labels)
        rij.append(coefficients)
        bi.append(offset)
    res_rij = numpy.array(rij, dtype=object)
    res_bi = numpy.array(bi, dtype=object)
    return res_rij, res_bi


def transform_string_to_digits(name: str, labels: Tuple[str]):
    """Transform string to digits.

    Multiplication has to be implicit, division must be explicit.
    White space within the string is optional.

    >>> transform_string_to_digits('-a+2x/7-0.7t+4', ('a', 'x', 't'))
    """
    coefficients = []
    offset = Fraction(0, 1).limit_denominator(10)
    l_name = name.strip().replace(" ", "").replace("+", " +").replace(
        "-", " -").split()
    for _label in labels:
        res = Fraction(0, 1).limit_denominator(10)
        for _name in l_name:
            if _label in _name:
                s_1 = _name.replace(_label, "").replace("+/", "+1/").replace(
                    "-/", "-1/")
                if s_1 == "":
                    s_1 = "1"
                if s_1.startswith("/"):
                    s_1 = "1" + s_1
                if s_1.endswith("+"):
                    s_1 = s_1 + "1"
                if s_1.endswith("-"):
                    s_1 = s_1 + "1"
                res += Fraction(s_1).limit_denominator(10)
        coefficients.append(res)
    res = Fraction(0, 1).limit_denominator(10)
    for _name in l_name:
        flag = all([not (_label in _name) for _label in labels])
        if flag:
            res += Fraction(_name).limit_denominator(10)
    offset = res
    return coefficients, offset


def transform_fraction_with_label_to_string(number: Fraction, label: str) \
        -> str:
    """Transform fraction with label to string.

    Parameters
    ----------
    number : Fraction
        DESCRIPTION.
    label : str
        DESCRIPTION.

    Returns
    -------
    str
        DESCRIPTION.

    """
    if isinstance(number, Fraction):
        val = number
    else:
        val = Fraction(number).limit_denominator(10)
    if val == Fraction(0, 1):
        res = ""
    elif val == Fraction(1, 1):
        res = f"{label:}"
    elif val == Fraction(-1, 1):
        res = f"-{label:}"
    elif val.denominator == 1:
        res = f"{val.numerator}{label:}"
    elif val.numerator == 1:
        res = f"{label:}/{val.denominator}"
    else:
        res = f"{val.numerator}{label:}/{val.denominator}"
    return res


def transform_digits_to_string(labels: Tuple[str], coefficients,
                               offset: Fraction) -> str:
    """Form a string from digits.

    Arguments
    ---------
        labels: the tuple of lablels (ex.: ('x', 'y', 'z') or ('a', 'b', 'c')))
        coefficients: the parameters in front of label (ex.: (1.0, 0.5, 0.0))
        offset: the number (ex.: 2/3)

    Output
    ------
    string

    Example
    -------
    >>> transform_digits_to_string(('x', 'y', 'z'), (1.0, 0.5, 0.0), 0.6666667)
    x+1/2y+2/3
    """
    l_res = []
    for _coefficient, _label in zip(coefficients, labels):
        _name = transform_fraction_with_label_to_string(_coefficient, _label)
        if _name == "":
            pass
        elif _name.startswith("-"):
            l_res.append(_name)
        elif l_res == []:
            l_res.append(_name)
        else:
            l_res.append(f"+{_name:}")
    _name = str(Fraction(offset).limit_denominator(10))
    if _name == "0":
        if l_res == []:
            l_res.append(_name)
    elif ((l_res == []) | (_name.startswith("-"))):
        l_res.append(_name)
    else:
        l_res.append(f"+{_name:}")
    return "".join(l_res)


def transform_r_b_to_string(r, b, labels=("x", "y", "z")) -> str:
    """Transform r (matrix) b (vector) to string.

    Parameters
    ----------
    r : TYPE
        DESCRIPTION.
    b : TYPE
        DESCRIPTION.
    labels : TYPE, optional
        DESCRIPTION. The default is ("x", "y", "z").

    Returns
    -------
    str
        DESCRIPTION.

    """
    l_res = [transform_digits_to_string(labels, _ri, _bi) for _ri, _bi in
             zip(r, b)]
    return ",".join(l_res)


def calc_mRmCmRT(r_ij, c_ij):
    """Calculate matrix multiplication R*C*RT.

    Matrices are expressed through its component and can be expressed as
    nD-array.

    r_ij: r11, r12, r13, r21, r22, r23, r31, r32, r33
    c_ij: c11, c12, c13, c21, c22, c23, c31, c32, c33
    """
    r11, r12, r13, r21, r22, r23, r31, r32, r33 = r_ij
    rt_ij = r11, r21, r31, r12, r22, r32, r13, r23, r33
    rcrt_ij = calc_product_matrices(r_ij, c_ij, rt_ij)
    rcrt_11, rcrt_12, rcrt_13, rcrt_21, rcrt_22, rcrt_23, rcrt_31, rcrt_32, \
        rcrt_33 = rcrt_ij
    return rcrt_11, rcrt_12, rcrt_13, rcrt_21, rcrt_22, rcrt_23, rcrt_31, \
        rcrt_32, rcrt_33


def calc_rotation_matrix_ij_by_euler_angles(alpha, beta, gamma):
    """Calculate rotational matrix from Euler angle.

    Tait-Bryan convection for angles is used.
    psi - x-axis
    theta - y axis
    phi - z axis
        0 <= psi   <= 2 pi ???
    -pi/2 <= theta <= pi/2
        0 <= phi   <= 2 pi
    """
    psi, theta, phi = alpha, beta, gamma
    rm_11 = numpy.cos(theta)*numpy.cos(phi)
    rm_21 = numpy.cos(theta)*numpy.sin(phi)
    rm_31 = -numpy.sin(theta)
    rm_12 = numpy.sin(psi)*numpy.sin(theta)*numpy.cos(phi) - \
        numpy.cos(psi)*numpy.sin(phi)
    rm_22 = numpy.sin(psi)*numpy.sin(theta)*numpy.sin(phi) + \
        numpy.cos(psi)*numpy.cos(phi)
    rm_32 = numpy.sin(psi)*numpy.cos(theta)
    rm_13 = numpy.cos(psi)*numpy.sin(theta)*numpy.cos(phi) + \
        numpy.sin(psi)*numpy.sin(phi)
    rm_23 = numpy.cos(psi)*numpy.sin(theta)*numpy.sin(phi) - \
        numpy.sin(psi)*numpy.cos(phi)
    rm_33 = numpy.cos(psi)*numpy.cos(theta)
    return rm_11, rm_12, rm_13, rm_21, rm_22, rm_23, rm_31, rm_32, rm_33


def calc_euler_angles_by_rotation_matrix_ij(rm_ij):
    """Calculate Euler Angles from rotational matrix.

    Tait-Bryan convection for angles used.
    psi - x-axis
    theta - y axis
    phi - z axis.
    """
    rm_11, rm_12, rm_13, rm_21, rm_22, rm_23, rm_31, rm_32, rm_33 = rm_ij
    if (abs(rm_31) != 1.):
        theta1 = -1*numpy.arcsin(rm_31)
        theta2 = numpy.pi-theta1
        psi1 = numpy.arctan2((rm_32)/numpy.cos(theta1),
                             (rm_33)/numpy.cos(theta1))
        psi2 = numpy.arctan2((rm_32)/numpy.cos(theta2),
                             (rm_33)/numpy.cos(theta2))
        phi1 = numpy.arctan2((rm_21)/numpy.cos(theta1),
                             (rm_11)/numpy.cos(theta1))
        phi2 = numpy.arctan2((rm_21)/numpy.cos(theta2),
                             (rm_11)/numpy.cos(theta2))
        res = [[psi1, theta1, phi1], [psi2, theta2, phi2]]
    else:
        phi = 0
        if (rm_31 == -1):
            theta = numpy.pi/2
            psi = phi+numpy.arctan2(rm_12, rm_13)
            res = [psi, theta, phi]
        else:
            theta = -1*numpy.pi/2
            psi = -1*phi+numpy.arctan2(-1*rm_12, -1*rm_13)
            res = [psi, theta, phi]
    return res


def calc_determinant_matrix_ij(m_ij):
    """Calculate determinant of the matrix / matrices.

    Parameters
    ----------
    m_ij : TYPE
        m_ij = m_11, m_12, m_13, m_21, m_22, m_23, m_31, m_32, m_33

    Returns
    -------
    det : TYPE
        DESCRIPTION.

    """
    m_11, m_12, m_13, m_21, m_22, m_23, m_31, m_32, m_33 = m_ij
    det = m_11*m_22*m_33 - m_11*m_23*m_32  \
        - m_12*m_21*m_33 + m_12*m_23*m_31  \
        + m_13*m_21*m_32 - m_13*m_22*m_31
    return det


def calc_inverse_matrix_ij(m_ij):
    """Calculate inverse matrix / matrices.

    Parameters
    ----------
    m_ij : TYPE
        m_ij = m_11, m_12, m_13, m_21, m_22, m_23, m_31, m_32, m_33

    Returns
    -------
    m_i_11 : TYPE
        DESCRIPTION.
    m_i_12 : TYPE
        DESCRIPTION.
    m_i_13 : TYPE
        DESCRIPTION.
    m_i_21 : TYPE
        DESCRIPTION.
    m_i_22 : TYPE
        DESCRIPTION.
    m_i_23 : TYPE
        DESCRIPTION.
    m_i_31 : TYPE
        DESCRIPTION.
    m_i_32 : TYPE
        DESCRIPTION.
    m_i_33 : TYPE
        DESCRIPTION.

    """
    eps = 1.0e-12
    det = calc_determinant_matrix_ij(m_ij)
    m_11, m_12, m_13, m_21, m_22, m_23, m_31, m_32, m_33 = m_ij
    inv_det = numpy.where(det < eps, 0., 1./det)
    m_i_11 = +(m_22*m_33-m_23*m_32) * inv_det
    m_i_21 = -(m_21*m_33-m_23*m_31) * inv_det
    m_i_31 = +(m_21*m_32-m_22*m_31) * inv_det
    m_i_12 = -(m_12*m_33-m_13*m_32) * inv_det
    m_i_22 = +(m_11*m_33-m_13*m_31) * inv_det
    m_i_32 = -(m_11*m_32-m_12*m_31) * inv_det
    m_i_13 = +(m_12*m_23-m_13*m_22) * inv_det
    m_i_23 = -(m_11*m_23-m_13*m_21) * inv_det
    m_i_33 = +(m_11*m_22-m_12*m_21) * inv_det
    return m_i_11, m_i_12, m_i_13, m_i_21, m_i_22, m_i_23, m_i_31, m_i_32, \
        m_i_33


def calc_rotation_matrix_ij_around_axis(angle, axis="x"):
    """Calculate rotation matrix around given axis.

    Parameters
    ----------
    angle : TYPE
        DESCRIPTION.
    axis : TYPE, optional
        DESCRIPTION. The default is "x".

    Returns
    -------
    r_m_11 : TYPE
        DESCRIPTION.
    r_m_12 : TYPE
        DESCRIPTION.
    r_m_13 : TYPE
        DESCRIPTION.
    r_m_21 : TYPE
        DESCRIPTION.
    r_m_22 : TYPE
        DESCRIPTION.
    r_m_23 : TYPE
        DESCRIPTION.
    r_m_31 : TYPE
        DESCRIPTION.
    r_m_32 : TYPE
        DESCRIPTION.
    r_m_33 : TYPE
        DESCRIPTION.

    """
    np_zero = 0.*angle
    np_one = 1. + np_zero
    if axis.lower() == "y":
        r_m_11 = numpy.cos(angle)
        r_m_12 = np_zero
        r_m_13 = numpy.sin(angle)
        r_m_21 = np_zero
        r_m_22 = np_one
        r_m_23 = np_zero
        r_m_31 = -numpy.sin(angle)
        r_m_32 = np_zero
        r_m_33 = numpy.cos(angle)
    elif axis.lower() == "z":
        r_m_11 = numpy.cos(angle)
        r_m_12 = numpy.sin(angle)
        r_m_13 = np_zero
        r_m_21 = -numpy.sin(angle)
        r_m_22 = numpy.cos(angle)
        r_m_23 = np_zero
        r_m_31 = np_zero
        r_m_32 = np_zero
        r_m_33 = np_one
    else:
        # by default "x"
        r_m_11 = np_one
        r_m_12 = np_zero
        r_m_13 = np_zero
        r_m_21 = np_zero
        r_m_22 = numpy.cos(angle)
        r_m_23 = numpy.sin(angle)
        r_m_31 = np_zero
        r_m_32 = -numpy.sin(angle)
        r_m_33 = numpy.cos(angle)
    return r_m_11, r_m_12, r_m_13, r_m_21, r_m_22, r_m_23, r_m_31, r_m_32, \
        r_m_33


def calc_product_matrices(m_1_ij, m_2_ij, *argv):
    """Product of two or more matrices."""
    if len(argv) == 0:
        m_1_11, m_1_12, m_1_13, m_1_21, m_1_22, m_1_23, m_1_31, m_1_32, \
            m_1_33 = m_1_ij
        m_2_11, m_2_12, m_2_13, m_2_21, m_2_22, m_2_23, m_2_31, m_2_32, \
            m_2_33 = m_2_ij

        m_3_11 = m_1_11*m_2_11 + m_1_12*m_2_21 + m_1_13*m_2_31
        m_3_12 = m_1_11*m_2_12 + m_1_12*m_2_22 + m_1_13*m_2_32
        m_3_13 = m_1_11*m_2_13 + m_1_12*m_2_23 + m_1_13*m_2_33

        m_3_21 = m_1_21*m_2_11 + m_1_22*m_2_21 + m_1_23*m_2_31
        m_3_22 = m_1_21*m_2_12 + m_1_22*m_2_22 + m_1_23*m_2_32
        m_3_23 = m_1_21*m_2_13 + m_1_22*m_2_23 + m_1_23*m_2_33

        m_3_31 = m_1_31*m_2_11 + m_1_32*m_2_21 + m_1_33*m_2_31
        m_3_32 = m_1_31*m_2_12 + m_1_32*m_2_22 + m_1_33*m_2_32
        m_3_33 = m_1_31*m_2_13 + m_1_32*m_2_23 + m_1_33*m_2_33
    else:
        m_h_ij = calc_product_matrices(m_1_ij, m_2_ij)
        m_3_11, m_3_12, m_3_13, m_3_21, m_3_22, m_3_23, m_3_31, m_3_32, \
            m_3_33 = calc_product_matrices(m_h_ij, *argv)
    return m_3_11, m_3_12, m_3_13, m_3_21, m_3_22, m_3_23, m_3_31, m_3_32, \
        m_3_33


def calc_product_matrix_vector(m_ij, v_j):
    """Product matrix and vector."""
    m_11, m_12, m_13, m_21, m_22, m_23, m_31, m_32, m_33 = m_ij
    v_1, v_2, v_3 = v_j
    o_1 = m_11*v_1 + m_12*v_2 + m_13*v_3
    o_2 = m_21*v_1 + m_22*v_2 + m_23*v_3
    o_3 = m_31*v_1 + m_32*v_2 + m_33*v_3
    return o_1, o_2, o_3


def calc_vector_angle(v_1_i, v_2_i):
    """Angle between two vectors (in radians)."""
    eps = 1e-12
    v_1_1, v_1_2, v_1_3 = v_1_i
    v_2_1, v_2_2, v_2_3 = v_2_i
    prod_v_12 = v_1_1*v_2_1 + v_1_2*v_2_2 + v_1_3*v_2_3
    norm_v_1 = numpy.sqrt(numpy.square(v_1_1) + numpy.square(v_1_2) +
                          numpy.square(v_1_3))
    norm_v_2 = numpy.sqrt(numpy.square(v_2_1) + numpy.square(v_2_2) +
                          numpy.square(v_2_3))
    flag_12 = numpy.logical_or((norm_v_1 > eps), (norm_v_2 > eps))

    angle = numpy.arccos(numpy.where(flag_12, prod_v_12 / (norm_v_1*norm_v_2),
                                     0.99999))
    return angle


def calc_vector_product(v_1_i, v_2_i):
    """Vector product of vectors."""
    v_1_1, v_1_2, v_1_3 = v_1_i
    v_2_1, v_2_2, v_2_3 = v_2_i
    o_1 = v_1_2 * v_2_3 - v_1_3 * v_2_2
    o_2 = v_1_3 * v_2_1 - v_1_1 * v_2_3
    o_3 = v_1_1 * v_2_2 - v_1_2 * v_2_1
    return o_1, o_2, o_3


def scalar_product(v_1_i, v_2_i):
    """Scalar product of vectors."""
    v_1_1, v_1_2, v_1_3 = v_1_i
    v_2_1, v_2_2, v_2_3 = v_2_i
    return v_1_1*v_2_1 + v_1_2*v_2_2 + v_1_3*v_2_3


def calc_rotation_matrix_by_two_vectors(v_1_i, v_2_i, ax_1="x", ax_2="y"):
    """Calculate rotation matrix by two vectors."""
    norm_v_1 = numpy.sqrt(numpy.square(v_1_i[0]) + numpy.square(v_1_i[1]) +
                          numpy.square(v_1_i[2]))
    norm_v_2 = numpy.sqrt(numpy.square(v_2_i[0]) + numpy.square(v_2_i[1]) +
                          numpy.square(v_2_i[2]))
    e_x = (v_1_i[0]/norm_v_1, v_1_i[1]/norm_v_1, v_1_i[2]/norm_v_1)
    e_12 = (v_2_i[0]/norm_v_2, v_2_i[1]/norm_v_2, v_2_i[2]/norm_v_2)
    v_z = calc_vector_product(e_x, e_12)
    norm_v_z = numpy.sqrt(numpy.square(v_z[0]) + numpy.square(v_z[1]) +
                          numpy.square(v_z[2]))
    e_z = (v_z[0]/norm_v_z, v_z[1]/norm_v_z, v_z[2]/norm_v_z)
    e_y = calc_vector_product(e_z, e_x)
    rm_11, rm_12, rm_13 = e_x[0], e_y[0], e_z[0]
    rm_21, rm_22, rm_23 = e_x[1], e_y[1], e_z[1]
    rm_31, rm_32, rm_33 = e_x[2], e_y[2], e_z[2]
    return rm_11, rm_12, rm_13, rm_21, rm_22, rm_23, rm_31, rm_32, rm_33


def tri_linear_interpolation(ind_xyz, den_3d):
    """Linear interpolation by four points in 3d array.

    Parameters
    ----------
    ind_xyz : [:, 3] of float
        DESCRIPTION.
    den_3d : TYPE
        DESCRIPTION.

    Returns
    -------
    c : TYPE
        DESCRIPTION.

    """
    ind_x, ind_y, ind_z = ind_xyz[:, 0], ind_xyz[:, 1], ind_xyz[:, 2]
    (n_x, n_y, n_z) = den_3d.shape
    ind_x0 = (numpy.floor(ind_x)).astype(int)
    ind_x1 = ind_x0+1
    ind_y0 = (numpy.floor(ind_y)).astype(int)
    ind_y1 = ind_y0+1
    ind_z0 = (numpy.floor(ind_z)).astype(int)
    ind_z1 = ind_z0+1

    xd = (ind_x-ind_x0)/(ind_x1-ind_x0)
    yd = (ind_y-ind_y0)/(ind_y1-ind_y0)
    zd = (ind_z-ind_z0)/(ind_z1-ind_z0)

    Vx0y0z0 = den_3d[numpy.mod(ind_x0, n_x), numpy.mod(ind_y0, n_y),
                     numpy.mod(ind_z0, n_z)]
    Vx1y0z0 = den_3d[numpy.mod(ind_x1, n_x), numpy.mod(ind_y0, n_y),
                     numpy.mod(ind_z0, n_z)]
    Vx0y0z1 = den_3d[numpy.mod(ind_x0, n_x), numpy.mod(ind_y0, n_y),
                     numpy.mod(ind_z1, n_z)]
    Vx1y0z1 = den_3d[numpy.mod(ind_x1, n_x), numpy.mod(ind_y0, n_y),
                     numpy.mod(ind_z1, n_z)]
    Vx0y1z0 = den_3d[numpy.mod(ind_x0, n_x), numpy.mod(ind_y1, n_y),
                     numpy.mod(ind_z0, n_z)]
    Vx1y1z0 = den_3d[numpy.mod(ind_x1, n_x), numpy.mod(ind_y1, n_y),
                     numpy.mod(ind_z0, n_z)]
    Vx0y1z1 = den_3d[numpy.mod(ind_x0, n_x), numpy.mod(ind_y1, n_y),
                     numpy.mod(ind_z1, n_z)]
    Vx1y1z1 = den_3d[numpy.mod(ind_x1, n_x), numpy.mod(ind_y1, n_y),
                     numpy.mod(ind_z1, n_z)]

    c00 = Vx0y0z0*(1.-xd) + Vx1y0z0*xd
    c01 = Vx0y0z1*(1.-xd) + Vx1y0z1*xd
    c10 = Vx0y1z0*(1.-xd) + Vx1y1z0*xd
    c11 = Vx0y1z1*(1.-xd) + Vx1y1z1*xd

    c0 = c00*(1.-yd) + c10*yd
    c1 = c01*(1.-yd) + c11*yd

    c = c0*(1.-zd) + c1*zd

    return c


def ortogonalize_matrix(m_ij, m_norm_ij):
    """Ortogonalize matrix.

    matrix m_ij is defined in coordinate system (a, b, c).
    It is given as tuple

    Matrix m_norm  used to recalculate coordinates  from direct space
    (a1/|a1|, a2/|a2|, a3/|a3|) to Cartesian one (x||a*, z||c).

    x_cart = m_norm * x_direct

    m_norm  = [[(1 - cos**2 alpha1 - cos**2 alpha2 - cos**2 alpha3 +
             2 cos alpha1 cos alpha2 cos alpha3)**0.5/sin(alpha1),  0,  0],
     [  (cos alpha3 - cos alpha1 cos alpha2) / sin alpha1, sin alpha1,  0],
     [                                         cos alpha2, cos alpha1,  1]]

    Should be given as

    m_norm_ij = (_11, _12, _13, _21, _22, _23, _31, _32, _33)

    Output matrix s_ij is defined in Cartezian coordinate system where
    x||a*, z||c, y= [z x] (right handed)
    """
    s_11, s_12, s_13, s_21, s_22, s_23, s_31, s_32, s_33 = calc_mRmCmRT(
        m_norm_ij, m_ij)
    return s_11, s_12, s_13, s_21, s_22, s_23, s_31, s_32, s_33


def calc_moment_2d_by_susceptibility(r_ij, susc_i, m_norm_ij, h_loc):
    """Calculate  moment 2d by susceptibility.

    Recalculate chi_i given in reciprocal unit cell according to symmetry
    elements for each point.

    After susceptibility is multiplied in magnetic field defined

    r_ij:= r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33
    susc_i:= chi_11, chi_22, chi_33, chi_12, chi_13, chi_23

    Matrix m_norm  used to recalculate coordinates  from direct space
    (a1/|a1|, a2/|a2|, a3/|a3|) to Cartesian one (x||a*, z||c).

    x_cart = m_norm * x_direct

    m_norm  = [[(1 - cos**2 alpha1 - cos**2 alpha2 - cos**2 alpha3 +
                 2 cos alpha1 cos alpha2 cos alpha3)**0.5/sin(alpha1),  0,  0],
         [  (cos alpha3 - cos alpha1 cos alpha2) / sin alpha1, sin alpha1,  0],
         [                                         cos alpha2, cos alpha1,  1]]

    Matrix m_norm should be given as

        m_norm_ij = (_11, _12, _13, _21, _22, _23, _31, _32, _33)

    moment_2d: [points, symmetry]
    """
    r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33 = r_ij
    chi_11, chi_22, chi_33, chi_12, chi_13, chi_23 = susc_i
    # [ind, symm]
    chi_2d_ij = calc_mRmCmRT(
        (r_11[numpy.newaxis, :], r_12[numpy.newaxis, :],
         r_13[numpy.newaxis, :], r_21[numpy.newaxis, :],
         r_22[numpy.newaxis, :], r_23[numpy.newaxis, :],
         r_31[numpy.newaxis, :], r_32[numpy.newaxis, :],
         r_33[numpy.newaxis, :]),
        (chi_11[:, numpy.newaxis], chi_12[:, numpy.newaxis],
         chi_13[:, numpy.newaxis], chi_12[:, numpy.newaxis],
         chi_22[:, numpy.newaxis], chi_23[:, numpy.newaxis],
         chi_13[:, numpy.newaxis], chi_23[:, numpy.newaxis],
         chi_33[:, numpy.newaxis]))
    chi_orto_ij = ortogonalize_matrix(chi_2d_ij, m_norm_ij)
    moment_2d = calc_product_matrix_vector(chi_orto_ij, h_loc)
    return moment_2d


def calc_phase_3d(hkl, r_ij, b_i, fract_xyz):
    r"""Calculate phases over 3 dimensions (hkl, points, symmetry).

    fract_xyz.shape = (pointxs, 3)

    r_ij, b_i are elements of symmetry

    phase = exp(2\pi i (h*Rs*x + h*bs))

    phase_3d: [hkl, points, symmetry]
    """
    index_h, index_k, index_l = hkl[0], hkl[1], hkl[2]
    x, y, z = fract_xyz[:, 0], fract_xyz[:, 1], fract_xyz[:, 2]
    r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33 = r_ij
    b_1, b_2, b_3 = b_i

    # [ind, symm]
    mr_2d_1, mr_2d_2, mr_2d_3 = calc_product_matrix_vector(
        (r_11[numpy.newaxis, :], r_12[numpy.newaxis, :],
         r_13[numpy.newaxis, :], r_21[numpy.newaxis, :],
         r_22[numpy.newaxis, :], r_23[numpy.newaxis, :],
         r_31[numpy.newaxis, :], r_32[numpy.newaxis, :],
         r_33[numpy.newaxis, :]),
        (x[:, numpy.newaxis], y[:, numpy.newaxis], z[:, numpy.newaxis]))

    # [hkl, ind, symm]
    hrr_3d = scalar_product(
        (index_h[:, numpy.newaxis, numpy.newaxis],
         index_k[:, numpy.newaxis, numpy.newaxis],
         index_l[:, numpy.newaxis, numpy.newaxis]),
        (mr_2d_1[numpy.newaxis, :, :], mr_2d_2[numpy.newaxis, :, :],
         mr_2d_3[numpy.newaxis, :, :]))

    # [hkl, symm]
    hb_2d = scalar_product(
        (index_h[:, numpy.newaxis], index_k[:, numpy.newaxis],
         index_l[:, numpy.newaxis]),
        (b_1[numpy.newaxis, :], b_2[numpy.newaxis, :], b_3[numpy.newaxis, :]))

    phase_3d = numpy.exp(2.*numpy.pi*1j*(hrr_3d+hb_2d[:, numpy.newaxis, :]))
    return phase_3d



from typing import NoReturn
import numpy
import matplotlib.pyplot as plt

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class TOFProc(ItemN):
    """Calculated intensity for time-of-flight experiment.

    Attributes
    ----------
        - time_corrected, d_spacing, intensity_plus_net,
          intensity_minus_net, intensity_plus_total, intensity_minus_total, 
          intensity_bkg_calc, intensity_net, intensity_total,
          intensity_diff_total, intensity_plus, intensity_plus_sigma,
          intensity_minus, intensity_minus_sigma, intensity,
          intensity_sigma (optional)

    time is given in microseconds.
    """
    ATTR_MANDATORY_NAMES = ("time", )
    ATTR_MANDATORY_TYPES = (float, )
    ATTR_MANDATORY_CIF = ("time", )

    ATTR_OPTIONAL_NAMES = (
        "time_corrected", "d_spacing", "intensity_plus_net",
        "intensity_minus_net", "intensity_plus_total", "intensity_minus_total", 
        "intensity_bkg_calc", "intensity_net", "intensity_total",
        "intensity_diff_total", "intensity_plus", "intensity_plus_sigma",
        "intensity_minus", "intensity_minus_sigma", "intensity",
        "intensity_sigma", "excluded")
    ATTR_OPTIONAL_TYPES = (float, float, float, float, float, float, float,
                           float, float, float, float, float, float, float, 
                           float, float, bool)
    ATTR_OPTIONAL_CIF = (
        "time_corrected", "d_spacing", "intensity_plus_net",
        "intensity_minus_net", "intensity_plus_total", "intensity_minus_total", 
        "intensity_bkg_calc", "intensity_net", "intensity_total",
        "intensity_diff_total", "intensity_plus", "intensity_plus_sigma",
        "intensity_minus", "intensity_minus_sigma", "intensity",
        "intensity_sigma", "excluded")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {
        "time_corrected": "{:.2f}", "d_spacing": "{:.5f}",
        "intensity_plus_net": "{:.2f}", "intensity_minus_net": "{:.2f}",
        "intensity_plus_total": "{:.2f}", "intensity_minus_total": "{:.2f}",
        "intensity_bkg_calc": "{:.2f}", "intensity_net": "{:.2f}",
        "intensity_total": "{:.2f}", "intensity_diff_total": "{:.2f}",
        "intensity_minus": "{:.2f}", "intensity_minus_sigma": "{:.2f}",
        "intensity": "{:.2f}", "intensity_sigma": "{:.2f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {"excluded": False}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "tof_proc"

    def __init__(self, **kwargs) -> NoReturn:
        super(TOFProc, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


class TOFProcL(LoopN):
    """Calculated intensity for time-of-flight experiment.
    """
    ITEM_CLASS = TOFProc
    ATTR_INDEX = "time"
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(TOFProcL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name


    def plots(self):
        return [self.plot_sum_diff()]
    
    def plot_sum_diff(self):
        """Plot unpolarized intensity vs. 2 theta
        """

        if (self.is_attribute("time") and self.is_attribute("intensity_plus") and 
                self.is_attribute("intensity_plus_sigma") and
                self.is_attribute("intensity_minus") and 
                self.is_attribute("intensity_minus_sigma") and
                self.is_attribute("intensity_plus_net") and
                self.is_attribute("intensity_minus_net") and
                self.is_attribute("intensity_bkg_calc")):

            fig = plt.figure(constrained_layout=False)
            gs = fig.add_gridspec(nrows=3, ncols=1, hspace=0, height_ratios=[7,1,7])

            ax_1 = fig.add_subplot(gs[0, 0])
            ax_3 = fig.add_subplot(gs[1, 0], sharex=ax_1)
            ax_3.set_yticks([])
            ax_2 = fig.add_subplot(gs[2, 0], sharex=ax_1)

            # fig, axs = plt.subplots(nrows = 2, sharex=True)
            # ax_1, ax_2 = axs
            ax_2.set_xlabel(r"time (ms)")
            ax_2.set_ylabel('Intensity (arb.u.)')
            ax_1.set_ylabel('Intensity (arb.u.)')

            np_excl = numpy.array(self.excluded, dtype=bool)
            np_notexcl = numpy.logical_not(np_excl)
            np_tth = numpy.array(self.time, dtype=float)
            np_up = numpy.array(self.intensity_plus, dtype=float)
            np_sup = numpy.array(self.intensity_plus_sigma, dtype=float)
            np_up_mod = numpy.array(self.intensity_plus_net, dtype=float)
            np_down = numpy.array(self.intensity_minus, dtype=float)
            np_sdown = numpy.array(self.intensity_minus_sigma, dtype=float)
            np_down_mod = numpy.array(self.intensity_minus_net, dtype=float)
            np_bkg = numpy.array(self.intensity_bkg_calc, dtype=float)
            np_sum = np_up + np_down 
            np_sum_mod = np_up_mod + np_down_mod + np_bkg
            np_diff = np_up - np_down 
            np_diff_mod = np_up_mod - np_down_mod
            np_ssum = numpy.sqrt(numpy.square(np_sup)+numpy.square(np_sdown))
            chi_sq_points = numpy.nansum(numpy.square((np_sum - np_sum_mod)/np_ssum)[np_notexcl])/numpy.sum(np_notexcl) 
            chi_sq_diff_points = numpy.nansum(numpy.square((np_diff - np_diff_mod)/np_ssum))/np_diff_mod.size

            ax_1.set_title(f"Unpolarized ($\chi^2/n = ${chi_sq_points:.2f}) and polarized ($\chi^2/n = ${chi_sq_diff_points:.2f}) signals")
            ax_1.fill_between(np_tth, (np_sum-np_ssum), (np_sum+np_ssum), where=np_notexcl, color="k", alpha=0.4, label="experiment")
            if numpy.any(np_excl):
                ax_1.fill_between(np_tth, (np_sum-np_ssum), (np_sum+np_ssum), where=np_excl, color="r", alpha=0.5, label="excluded")

            # ax_1.errorbar(np_tth[np_notexcl], np_sum[np_notexcl], yerr=np_ssum[np_notexcl], fmt="ko", alpha=0.2, label="experiment")
            # ax_1.errorbar(np_tth[np_excl], np_sum[np_excl], yerr=np_ssum[np_excl], fmt="rs", alpha=0.2, label="excluded")

            y_min_d, y_max_d = ax_1.get_ylim()
            param = y_min_d-((np_sum - np_sum_mod)[np_notexcl]).max()
            coeff = np_notexcl.astype(int)
            ax_1.plot([np_tth.min(), np_tth.max()], [param, param], "k:")
            ax_1.plot(np_tth, coeff*(np_sum - np_sum_mod)+param, "r-", alpha=0.5, label="difference")
            ax_1.plot(np_tth, np_sum_mod, "b-", label="model", linewidth=2)
            ax_1.plot(np_tth, np_bkg, "b:", label="background")
            ax_1.legend(loc='upper right')

            ax_2.plot([np_tth.min(), np_tth.max()], [0., 0.], "b:")
            ax_2.errorbar(np_tth, np_diff, yerr=np_ssum, fmt="ko", alpha=0.2, label="experiment")

            y_min_d, y_max_d = ax_2.get_ylim()
            param = y_min_d-(np_diff-np_diff_mod).max()
            ax_2.plot([np_tth.min(), np_tth.max()], [param, param], "k:")
            ax_2.plot(np_tth, np_diff-np_diff_mod+param, "r-", alpha=0.5, label="difference")
            ax_2.plot(np_tth, np_diff_mod, "b-", label="model", linewidth=2)
            ax_2.legend(loc='upper right')

        elif (self.is_attribute("time") and 
                self.is_attribute("intensity") and
                self.is_attribute("intensity_plus_net") and
                self.is_attribute("intensity_minus_net") and
                self.is_attribute("intensity_bkg_calc")):

            fig = plt.figure(constrained_layout=False)
            gs = fig.add_gridspec(nrows=2, ncols=1, hspace=0, height_ratios=[14,1])

            ax_1 = fig.add_subplot(gs[0, 0])
            ax_3 = fig.add_subplot(gs[1, 0], sharex=ax_1)
            ax_3.set_yticks([])

            ax_1.set_xlabel(r"time (ms)")
            ax_1.set_ylabel('Intensity (arb.u.)')

            np_excl = numpy.array(self.excluded, dtype=bool)
            np_notexcl = numpy.logical_not(np_excl)
            np_tth = numpy.array(self.time, dtype=float)
            np_sum = numpy.array(self.intensity, dtype=float)
            np_ssum = numpy.array(self.intensity_sigma, dtype=float)

            np_up_mod = numpy.array(self.intensity_plus_net, dtype=float)
            np_down_mod = numpy.array(self.intensity_minus_net, dtype=float)
            np_bkg = numpy.array(self.intensity_bkg_calc, dtype=float)
            np_sum_mod = np_up_mod + np_down_mod + np_bkg

            chi_sq_points = numpy.nansum(numpy.square((np_sum - np_sum_mod)/np_ssum)[np_notexcl])/numpy.sum(np_notexcl) 
            ax_1.set_title(f"Unpolarized signal $\chi^2/n = ${chi_sq_points:.2f}")

            ax_1.fill_between(np_tth, (np_sum-np_ssum), (np_sum+np_ssum), where=np_notexcl, color="k", alpha=0.4, label="experiment")
            ax_1.fill_between(np_tth, (np_sum-np_ssum), (np_sum+np_ssum), where=np_excl, color="r", alpha=0.5, label="excluded")

            # ax_1.errorbar(np_tth[np_notexcl], np_sum[np_notexcl], yerr=np_ssum[np_notexcl], fmt="ko", alpha=0.2, label="experiment")
            # ax_1.errorbar(np_tth[np_excl], np_sum[np_excl], yerr=np_ssum[np_excl], fmt="rs", alpha=0.2, label="excluded")

            y_min_d, y_max_d = ax_1.get_ylim()
            param = y_min_d-(np_sum - np_sum_mod).max()
            coeff = np_notexcl.astype(int)

            ax_1.plot([np_tth.min(), np_tth.max()], [param, param], "k:")
            ax_1.plot(np_tth, coeff*(np_sum - np_sum_mod)+param, "r-", alpha=0.5, label="difference")
            ax_1.plot(np_tth, np_sum_mod, "b-", label="model", linewidth=2)

            if (self.is_attribute("ttheta") and
                    self.is_attribute("intensity_bkg_calc")):
                np_tth = numpy.array(self.ttheta, dtype=float)
                np_bkg = numpy.array(self.intensity_bkg_calc, dtype=float)
                ax_1.plot(np_tth, np_bkg, "b:", label="background")
            ax_1.legend(loc='upper right')
            
        fig.tight_layout()
        return (fig, ax_1)

    def estimate_background(self, tof_background):
        pass

# s_cont = """
#   loop_
#   _tof_proc_time
#   _tof_proc_time_corrected
#   _tof_proc_d_spacing
#   _tof_proc_intensity_plus_net
#   _tof_proc_intensity_minus_net
#   _tof_proc_intensity_plus_total
#   _tof_proc_intensity_minus_total
#   _tof_proc_intensity_bkg_calc
#   _tof_proc_intensity_plus
#   _tof_proc_intensity_plus_sigma
#   _tof_proc_intensity_minus
#   _tof_proc_intensity_minus_sigma
#   4.00  3.9  11.2   400.00   317.00   460.00   377.00   60.00    465.80000   128.97000   301.88000   129.30000"""

# obj = TOFProcL.from_cif(s_cont)
# print(obj, end="\n\n")



"""Functions to find roots."""
import math

def rootsearch(f, a: float, b: float, dx: float):
    """Seacrh one root in range [a, b]."""
    x1 = a
    f1 = f(a)
    x2 = a + dx
    f2 = f(x2)
    if x2 > b:
        x2 = b
        f2 = f(x2)
    while f1*f2 > 0.0:
        if x1 >= b:
            return None, None
        x1 = x2
        f1 = f2
        x2 = x1 + dx
        f2 = f(x2)
        if x2 > b:
            x2 = b
            f2 = f(x2)
    return x1, x2


def bisect(f, x1: float, x2: float, switch: int = 0, epsilon: float = 1.0e-12):
    """Bisecting."""
    f1 = f(x1)
    if f1 == 0.0:
        return x1
    f2 = f(x2)
    if f2 == 0.0:
        return x2
    if f1*f2 > 0.0:
        print('Root is not bracketed')
        return None
    n = int(math.ceil(math.log(abs(x2 - x1)/epsilon)/math.log(2.0)))
    for i in range(n):
        x3 = 0.5*(x1 + x2)
        f3 = f(x3)
        if (switch == 1) and (abs(f3) > abs(f1)) and (abs(f3) > abs(f2)):
            return None
        if f3 == 0.0:
            return x3
        if f2*f3 < 0.0:
            x1 = x3
            f1 = f3
        else:
            x2 = x3
            f2 = f3
    return (x1 + x2)/2.0


def calc_roots(f, a: float, b: float, eps: float = 0):
    """
    Find all roots.

    Function f is given in range [a,b].
    Roots should not be close than eps.

    Give also estimation of mistake if mistake of zero is defined.
    """
    if eps == 0:
        eps = abs(b-a)*1./100
    l_res = []
    x1, x2 = rootsearch(f, a, b, eps)
    while x1 is not None:
        a2 = x2
        root = bisect(f, x1, x2, 1)
        if root is not None:
            # res.append( (round(root,-int(math.log(eps, 10)))))
            l_res.append(root)
        x1, x2 = rootsearch(f, a2, b, eps)
    return l_res



"""Module realized some mathematical functions.

Functions
---------
    - calc_m_sigma
    - calc_scalar_product_by_vectors
    - calc_scalar_product_by_complex_vectors
    - calc_modulus_sq_by_complex_vector

"""
import numpy


def calc_m_sigma(np_M, np_sigma):
    """Estimate errors for equation.

    np_M * np_val * np_M^T
    """
    np_sig_1 = numpy.zeros((3, 3), dtype=float)
    for _i in range(3):
        for _j in range(3):
            res = []
            for _k in range(3):
                for _l in range(3):
                    res.append((np_sigma[_k, _l]*np_M[_i, _k]*np_M[_j, _l])**2)
            np_sig_1[_i, _j] = sum(res)
    return numpy.sqrt(np_sig_1)


def calc_scalar_product_by_vectors(vector_1, vector_2, flag_vector_1=False,
                                   flag_vector_2=False):
    """Scalar product by vectors.

    The vector is given as tuple of its coordinates defined in Chartezian
    coordinate system
    """
    product = tuple([v_1*v_2 for v_1, v_2 in zip(vector_1, vector_2)])
    flag_derivatives = any([flag_vector_1, flag_vector_2])
    if flag_derivatives:
        der_vector_1, der_vector_2 = [], []
        for v_1, v_2 in zip(vector_1, vector_2):
            der_vector_1.append((v_1*0.+1.)*v_2)
            der_vector_2.append(v_1*(v_2*0.+1.))
        dder = {"der__vector_1": tuple(der_vector_1),
                "der__vector_2": tuple(der_vector_2)}
    else:
        dder = {}
    return product, dder


def calc_scalar_product_by_complex_vectors(
        vector_1, vector_2, flag_vector_1=False, flag_vector_2=False):
    """Scalar product by complex vectors.

    The vector is given as tuple of its coordinates defined in Chartezian
    coordinate system.
    """
    product_comp = tuple([v_1*v_2 for v_1, v_2 in zip(vector_1, vector_2)])
    product = product_comp[0] + product_comp[1] + product_comp[2]
    flag_derivatives = any([flag_vector_1, flag_vector_2])
    dder = {}
    if flag_derivatives:
        der_vector_1_re, der_vector_1_im = [], []
        der_vector_2_re, der_vector_2_im = [], []
        for v_1, v_2 in zip(vector_1, vector_2):
            der_vector_1_re.append((v_1*0.+1.)*v_2)
            der_vector_1_im.append((v_1*0.+1.)*v_2)
            der_vector_2_re.append(v_1*(v_2*0.+1.))
            der_vector_2_im.append(v_1*(v_2*0.+1.))
        if flag_vector_1:
            dder["der__vector_1_re"] = tuple(der_vector_1_re)
            dder["der__vector_1_im"] = tuple(der_vector_1_im)
        if flag_vector_2:
            dder["der__vector_2_re"] = tuple(der_vector_2_re)
            dder["der__vector_2_im"] = tuple(der_vector_2_im)
    return product, dder


def calc_modulus_sq_by_complex_vector(vector_1, flag_vector_1=False):
    """Square of modulus of complex vector.

    The vector is given as tuple of its coordinates defined in Chartezian
    coordinate system.
    """
    modulus_comp = tuple([(v_1*v_1.conjugate()).real for v_1 in vector_1])
    modulus = modulus_comp[0]+modulus_comp[1]+modulus_comp[2]
    flag_derivatives = any([flag_vector_1])
    if flag_derivatives:
        der_vector_1_re, der_vector_1_im = [], []
        for v_1 in vector_1:
            der_vector_1_re.append(2.*v_1.real)
            der_vector_1_im.append(2.*v_1.imag)
        dder = {"der__vector_1_re": tuple(der_vector_1_re),
                "der__vector_1_im": tuple(der_vector_1_im)}
    else:
        dder = {}
    return modulus, dder



from typing import NoReturn
import numpy

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary
from cryspy.A_functions_base.function_1_strings import \
    transform_string_to_r_b

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class SpaceGroupSymopMagnCentering(ItemN):
    """Centring translation in description of a magnetic space group.
    
    Centering or anti-centering translation.

    Given in a BNS-supercell description of a magnetic space group.

    Attributes
    ----------
    - xyz (mandatory)
    - description, id (optioanl)
    """

    ATTR_MANDATORY_NAMES = ("xyz", )
    ATTR_MANDATORY_TYPES = (str, )
    ATTR_MANDATORY_CIF = ("xyz", )

    ATTR_OPTIONAL_NAMES = ("id", "description")
    ATTR_OPTIONAL_TYPES = (str, str)
    ATTR_OPTIONAL_CIF = ("id", "description")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ("r", "b", "r_11", "r_12", "r_13", "r_21", "r_22",
                      "r_23", "r_31", "r_32", "r_33", "b_1", "b_2", "b_3",
                      "theta", "sym_elem")
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "space_group_symop_magn_centering"

    def __init__(self, **kwargs) -> NoReturn:
        super(SpaceGroupSymopMagnCentering, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def form_object(self) -> NoReturn:
        """Form object."""
        xyz = self.xyz
        if xyz is None:
            return False
        r, b = transform_string_to_r_b(xyz, labels=("x", "y", "z"))
        r_11, r_12, r_13 = int(r[0, 0]), int(r[0, 1]), int(r[0, 2])
        r_21, r_22, r_23 = int(r[1, 0]), int(r[1, 1]), int(r[1, 2])
        r_31, r_32, r_33 = int(r[2, 0]), int(r[2, 1]), int(r[2, 2])
        den_1, den_2 = b[0].denominator, b[1].denominator
        den_3 = b[2].denominator
        num_1, num_2 = b[0].numerator, b[1].numerator
        num_3 = b[2].numerator
        den = numpy.lcm.reduce([den_1, den_2, den_3])
        num_1 *= den//den_1
        num_2 *= den//den_2
        num_3 *= den//den_3

        theta = int(b[3])
        self.__dict__["r_11"] = r_11
        self.__dict__["r_12"] = r_12
        self.__dict__["r_13"] = r_13
        self.__dict__["r_21"] = r_21
        self.__dict__["r_22"] = r_22
        self.__dict__["r_23"] = r_23
        self.__dict__["r_31"] = r_31
        self.__dict__["r_32"] = r_32
        self.__dict__["r_33"] = r_33
        self.__dict__["b_1"] = b[0]
        self.__dict__["b_2"] = b[1]
        self.__dict__["b_3"] = b[2]
        self.__dict__["theta"] = theta
        self.__dict__["sym_elem"] = numpy.array([
            num_1, num_2, num_3, den,
            r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33,
            theta], dtype=int)
        # self.__dict__["sym_elem"] = numpy.array([
        #     num_1, num_2, num_3, den,
        #     r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33,
        #     theta*r_11, theta*r_12, theta*r_13, theta*r_21, theta*r_22,
        #     theta*r_23, theta*r_31, theta*r_32, theta*r_33], dtype=int)


class SpaceGroupSymopMagnCenteringL(LoopN):
    """Centring translation in description of a magnetic space group.
    
     This loop provides a list of centering or anti-centering
     translation in a BNS-supercell description of a magnetic space
     group.

     Keeping the centering and anti-centering translations in a
     separate loop leaves only representative point operations in the
     main SPACE_GROUP_SYMOP_MAGN_OPERATION loop. The direct sum of
     the two loops produces the full set of representative operations
     of the magnetic space group.  This centering loop is optional, so
     that it is always possible to include all of the symmetry
     operations in the main loop.

     When this centering loop is employed, the representative point
     operations in the main SPACE_GROUP_SYMOP_MAGN_OPERATION loop may
     not form a closed subgroup, but instead  generate some of the
     fractional translations of the centering loop.  Despite this
     annoyance, a separate centering loop is important because
     magnetic structures tend to have a relatively large number of
     centering and anti-centering translations, which can make the
     resulting list of operators very long and unintuitive, especially
     when working  in non-standard settings.

     One could argue that anti-centering operations belong in the main
     representative- point-operation loop since they are not actually
     translations of the magnetic lattice.   In fact, a pure time
     reversal is a generator of the magnetic point group of a  type-4
     magnetic space group.  Nevertheless, this centering loop is
     defined to  include the anti-centerings due to the common
     practice of referring to a "black and white" lattice of
     centerings and anti-centerings."""

    ITEM_CLASS = SpaceGroupSymopMagnCentering
    ATTR_INDEX = "id"

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(SpaceGroupSymopMagnCenteringL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

    def get_sym_elems(self):
        """Get sym elems."""
        res = numpy.array([item.sym_elem for item in self.items], dtype=int
                          ).transpose()
        return res

# s_cont = """
# loop_
# _space_group_symop_magn_centering_id
# _space_group_symop_magn_centering_xyz
# 1   x,y,z,+1
# """

# obj = SpaceGroupSymopMagnCenteringL.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj["1"], end="\n\n")
# print(obj.get_sym_elems(), end="\n\n")
# print(obj.get_sym_elems().shape, end="\n\n")



import numpy
from typing import NoReturn
import scipy

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN

from cryspy.C_item_loop_classes.cl_1_pd_meas import PdMeasL

na = numpy.newaxis

class PdBackground(ItemN):
    """Background point in powder 1d experiment.
    """
    ATTR_MANDATORY_NAMES = ("ttheta", "intensity")
    ATTR_MANDATORY_TYPES = (float, float)
    ATTR_MANDATORY_CIF = ("2theta", "intensity")

    ATTR_OPTIONAL_NAMES = ()
    ATTR_OPTIONAL_TYPES = ()
    ATTR_OPTIONAL_CIF = ()

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("intensity", )
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "pd_background"

    def __init__(self, **kwargs) -> NoReturn:
        super(PdBackground, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"ttheta": 0}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


class PdBackgroundL(LoopN):
    """Background points in powder 1d experiment.
    """
    ITEM_CLASS = PdBackground
    ATTR_INDEX = "ttheta"
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(PdBackgroundL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

    def interpolate_by_points(self, tth):
        l_ttheta = self.ttheta
        l_intensity = self.intensity
        tth_b = numpy.array(l_ttheta, dtype=float)
        int_b = numpy.array(l_intensity, dtype=float)
        if len(l_ttheta) == 0:
            int_1d = numpy.zeros(tth.size, dtype=float)
        else:
            int_1d = numpy.interp(tth, tth_b, int_b)
        return int_1d

    def define_points(self, pd_meas: PdMeasL, step_ttheta: float = 10.):
        ttheta = numpy.array(pd_meas.ttheta, dtype=float)
        if pd_meas.items[0].is_attribute("intensity_plus"):
            intensity = (
                numpy.array(pd_meas.intensity_plus, dtype=float) +
                numpy.array(pd_meas.intensity_minus, dtype=float))
            intensity_sigma = numpy.sqrt(
                numpy.square(numpy.array(pd_meas.intensity_plus_sigma, dtype=float)) +
                numpy.square(numpy.array(pd_meas.intensity_minus_sigma, dtype=float)))
        else:
            intensity = numpy.array(pd_meas.intensity, dtype=float)
            intensity_sigma = numpy.array(pd_meas.intensity_sigma, dtype=float)

        step_n = int(step_ttheta/(ttheta[1]-ttheta[0]))
        if step_n == 0:
            step_n = 1
        intensity_bkg = estimate_background(intensity, step_n=step_n)

        delta_0 = 0.
        res = scipy.optimize.minimize(
            lambda p: calc_chi_sq(intensity_bkg, intensity, intensity_sigma, p),
            delta_0)
        delta_0 = res["x"]
        intensity_bkg += delta_0

        # n_points = int((ttheta.max()-ttheta.min())/step_ttheta + 2)
        # ttheta_bkgr = numpy.linspace(ttheta.min(), ttheta.max(), n_points, endpoint=True)
        #
        # flags = numpy.abs(ttheta[na, :]-ttheta_bkgr[:, na])<step_ttheta
        #
        # int_bkrg = numpy.array([numpy.nanmin(intensity_bkg[flag], axis=0) for flag in flags], dtype=float)
        # int_bkrg = numpy.round(int_bkrg, decimals=5)
        self.numpy_ttheta = ttheta
        self.numpy_intensity = intensity_bkg
        self.items = []
        self.numpy_to_items()
        return

    def get_dictionary(self):
        res = {}
        res["background_ttheta"] = numpy.array(self.ttheta, dtype=float)*numpy.pi/180.
        res["background_intensity"] = numpy.array(self.intensity, dtype=float)
        res["flags_background_intensity"] = numpy.array(self.intensity_refinement, dtype=bool)
        return res

def estimate_background(y_exp, step_n: int = 10):
    n_points = numpy.arange(1, step_n+1)
    y_aver = y_exp.sum()/y_exp.size
    y_min = y_exp.min()
    flag = y_exp < y_aver + 2*(y_aver-y_min)

    # y_bkg_es = numpy.zeros_like(y_exp)
    # y_bkg_es[flag] = y_exp[flag]
    # y_bkg_es[numpy.logical_not(flag)] = y_aver+2*(y_aver-y_min)
    y_bkg_es = numpy.copy(y_exp)

    n_total = y_bkg_es.size
    for iii in range(100):
        y_bkg_new = numpy.zeros_like(y_bkg_es)
        for i_point in range(n_total):
            ind_left = i_point - n_points
            ind_right = i_point + n_points

            flag_left = ind_left < 0
            if numpy.any(flag_left):
                ind_left[flag_left] = 0
                ind_right[flag_left] = 0

            flag_right = ind_right >= n_total
            if numpy.any(flag_right):
                ind_right[flag_right] = n_total-1
                ind_left[flag_right] = n_total-1

            y_left = y_bkg_es[ind_left]
            y_right = y_bkg_es[ind_right]

            y_bkg_new[i_point] = (y_left+y_right).sum()/(2*step_n)


        if iii != 99:
            flag = y_bkg_new > y_bkg_es
            y_bkg_new[flag] = y_bkg_es[flag]
        y_bkg_es = y_bkg_new
    y_bkg = y_bkg_es
    return y_bkg


def calc_chi_sq(y_bkg, y_exp, sy_exp, delta):
    y_diff = (y_exp-(y_bkg+delta))/sy_exp
    flag = y_diff > 1
    y_diff[flag] = 1.
    wss = numpy.sum(numpy.square(y_diff))
    return wss


"""
:Constants:


:Functions:



"""

from cryspy.A_functions_base.database import DATABASE
from cryspy.A_functions_base.charge_form_factor import get_atomic_symbol_ion_charge_isotope_number_by_ion_symbol


def apply_constraint_on_cell_by_type_cell(cell, type_cell:str,
                                          it_coordinate_system_code:str):
    length_a, length_b, length_c = cell.length_a, cell.length_b, cell.length_c
    angle_alpha, angle_beta = cell.angle_alpha, cell.angle_beta
    angle_gamma = cell.angle_gamma
    if length_a is None: 
        cell.length_a = 1.
    if length_b is None: 
        cell.length_b = 1.
    if length_c is None: 
        cell.length_c = 1.
    if angle_alpha is None: 
        cell.angle_alpha = 90.
    if angle_beta is None: 
        cell.angle_beta = 90.
    if angle_gamma is None: 
        cell.angle_gamma = 90.

    if type_cell == "aP":
        pass
    elif type_cell.startswith("m"):
        cell.angle_alpha, cell.angle_alpha_refinement, cell.angle_alpha_constraint = 90., False, True
        cell.angle_gamma, cell.angle_gamma_refinement, cell.angle_gamma_constraint = 90., False, True
    elif type_cell.startswith("o"):
        cell.angle_alpha, cell.angle_alpha_refinement, cell.angle_alpha_constraint = 90., False, True
        cell.angle_beta, cell.angle_beta_refinement, cell.angle_beta_constraint = 90., False, True
        cell.angle_gamma, cell.angle_gamma_refinement, cell.angle_gamma_constraint = 90., False, True
    elif ((type_cell.startswith("t")) | (type_cell == "hP")):
        cell.length_b, cell.length_b_sigma = cell.length_a, cell.length_a_sigma
        cell.length_b_refinement, cell.length_b_constraint = False, True
        cell.angle_alpha, cell.angle_alpha_refinement, cell.angle_alpha_constraint = 90., False, True
        cell.angle_beta, cell.angle_beta_refinement, cell.angle_beta_constraint = 90., False, True
        cell.angle_gamma, cell.angle_gamma_refinement, cell.angle_gamma_constraint = 90., False, True
    elif (type_cell == "hR"):
        if it_coordinate_system_code.lower() == "h":
            cell.length_b, cell.length_b_sigma = cell.length_a, cell.length_a_sigma
            cell.length_b_refinement, cell.length_b_constraint = False, True
            cell.angle_alpha, cell.angle_alpha_refinement, cell.angle_alpha_constraint = 90., False, True
            cell.angle_beta, cell.angle_beta_refinement, cell.angle_beta_constraint = 90., False, True
            cell.angle_gamma, cell.angle_gamma_refinement, cell.angle_gamma_constraint = 120., False, True
        else:
            cell.length_b, cell.length_b_sigma = cell.length_a, cell.length_a_sigma
            cell.length_b.refinement, cell.length_b_constraint = False, True
            cell.length_c, cell.length_c_sigma = cell.length_a, cell.length_a_sigma
            cell.length_c_refinement, cell.length_c_constraint = False, True
            cell.angle_beta, cell.angle_beta_sigma = cell.angle_alpha, cell.angle_alpha_sigma
            cell.angle_beta_refinement, cell.angle_beta_constraint = False, True
            cell.angle_gamma, cell.angle_gamma_sigma = cell.angle_alph, cell.angle_alpha_sigma
            cell.angle_gamma_refinement, cell.angle_gamma_constraint = False, True
    elif type_cell.startswith("c"):
        cell.length_b, cell.length_b_sigma = cell.length_a, cell.length_a_sigma
        cell.length_b.refinement, cell.length_b_constraint = False, True
        cell.length_c, cell.length_c_sigma = cell.length_a, cell.length_a_sigma
        cell.length_c_refinement, cell.length_c_constraint = False, True
        cell.angle_alpha, cell.angle_alpha_refinement, cell.angle_alpha_constraint = 90., False, True
        cell.angle_beta, cell.angle_beta_refinement, cell.angle_beta_constraint = 90., False, True
        cell.angle_gamma, cell.angle_gamma_refinement, cell.angle_gamma_constraint = 90., False, True


def get_scat_length_neutron(type_n: str):
    """
    Take scat_length_neutron.
    """
    atomic_symbol, ion_charge, isotope_number = get_atomic_symbol_ion_charge_isotope_number_by_ion_symbol(type_n)
    if isotope_number is None:
        s_isotope = atomic_symbol
    else:
        s_isotope = f"{isotope_number:}{atomic_symbol:}"
    res = DATABASE["Isotopes"][("b_scat", s_isotope)]
    return res



"""Calculation section from density_point.

Functions
---------
    - calc_section_from_density_point
"""
import numpy

from cryspy.A_functions_base.function_1_matrices import \
    tri_linear_interpolation

from cryspy.A_functions_base.function_2_mem import transfer_to_density_3d

from cryspy.C_item_loop_classes.cl_1_atom_site import AtomSiteL
from cryspy.C_item_loop_classes.cl_1_atom_site_susceptibility import \
    AtomSiteSusceptibilityL
from cryspy.C_item_loop_classes.cl_1_cell import Cell
from cryspy.C_item_loop_classes.cl_1_mem_parameters import MEMParameters
from cryspy.C_item_loop_classes.cl_1_space_group_symop import \
    SpaceGroupSymop, SpaceGroupSymopL

from cryspy.C_item_loop_classes.cl_2_section import Section

from cryspy.C_item_loop_classes.cl_3_density_point import DensityPoint, \
    DensityPointL

def read_file_den(file_name: str):
    """Read density file.
    
    Arguments
    ---------
        -   file_name is a file name
    
    Output
    ------
        - cell
        - space_group_symop
        - density_point
    """
    with open(file_name, "r") as fid:
        l_content = fid.readlines()

    number_lines = int(l_content[1])

    hh = l_content[number_lines+2].strip().split()
    cell = Cell(length_a=float(hh[0]), length_b=float(hh[1]),
                length_c=float(hh[2]), angle_alpha=float(hh[3]),
                angle_beta=float(hh[4]), angle_gamma=float(hh[5]))

    number_points = [int(hh) for hh in 
                     l_content[number_lines+3][:-1].split()[:3]]
    [n_el_symm, centr, n_orig] = [int(hh) for hh in 
                                  l_content[number_lines+3][:-1].split()[3:6]]
    
    l_item_dp = []
    for line in l_content[2:number_lines+2]:
        hh = line.strip().split()
        ind_x, ind_y, ind_z = int(hh[0]), int(hh[1]), int(hh[2])
        den, den_f, den_a = 0., 0., 0.
        if len(hh) == 6:
            den_f = float(hh[4])
            den_a = float(hh[5])
        else:
            den_u = float(hh[3])
            if den_u >= 0.:
                den_f = den_u
            else:
                den_a = den_u
        item = DensityPoint(
            index_x=ind_x, index_y=ind_y, index_z=ind_z,
            fract_x=float(ind_x)/float(number_points[0]),
            fract_y=float(ind_y)/float(number_points[1]),
            fract_z=float(ind_z)/float(number_points[2]),
            density=den, density_ferro=den_f, density_antiferro=den_a)
        l_item_dp.append(item)
    density_point = DensityPointL()
    density_point.items = l_item_dp


    l_el_symm = []
    for line in l_content[(number_lines+4):(number_lines+4+n_el_symm)]:
        hh = line[:-1].strip().split()
        l_el_symm.append([float(hh[9]), int(hh[0]), int(hh[1]), int(hh[2]),
                          float(hh[10]), int(hh[3]), int(hh[4]), int(hh[5]),
                          float(hh[11]), int(hh[6]), int(hh[7]), int(hh[8])])
    l_item_sg = []
    for line in l_content[(number_lines+4+n_el_symm):
                          (number_lines+4+n_el_symm+n_orig)]:
        orig = [float(hh) for hh in line[:-1].split()[:3]]
        for el_symm in l_el_symm:
            el_symm_full = [
                (el_symm[0]+orig[0])%1., el_symm[1], el_symm[2], el_symm[3],
                (el_symm[4]+orig[1])%1., el_symm[5], el_symm[6], el_symm[7],
                (el_symm[8]+orig[2])%1., el_symm[9], el_symm[10], el_symm[11]]
            item = SpaceGroupSymop.define_by_el_symm(el_symm_full)
            l_item_sg.append(item)
    space_group_symop = SpaceGroupSymopL()
    space_group_symop.items = l_item_sg
    return cell, space_group_symop, density_point, number_points

def calc_section_from_density_point(
        section: Section, density_point: DensityPointL,
        mem_parameters: MEMParameters, cell: Cell,
        space_group_symop: SpaceGroupSymopL, atom_site: AtomSiteL,
        atom_site_susceptibility: AtomSiteSusceptibilityL):
    """Calculate section in density point.

    Arguments
    ---------
        - section
        - density_point
        - cell
        - space_group_symop

    Output:
        - matrix_chi, matrix_b
    """
    r_11 = numpy.array(space_group_symop.r_11, dtype=int)
    r_12 = numpy.array(space_group_symop.r_12, dtype=int)
    r_13 = numpy.array(space_group_symop.r_13, dtype=int)
    r_21 = numpy.array(space_group_symop.r_21, dtype=int)
    r_22 = numpy.array(space_group_symop.r_22, dtype=int)
    r_23 = numpy.array(space_group_symop.r_23, dtype=int)
    r_31 = numpy.array(space_group_symop.r_31, dtype=int)
    r_32 = numpy.array(space_group_symop.r_32, dtype=int)
    r_33 = numpy.array(space_group_symop.r_33, dtype=int)

    b_1 = numpy.array(space_group_symop.b_1, dtype=float)
    b_2 = numpy.array(space_group_symop.b_2, dtype=float)
    b_3 = numpy.array(space_group_symop.b_3, dtype=float)

    r_ij = (r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33)
    b_i = (b_1, b_2, b_3)

    np_ind_x = density_point.numpy_index_x
    np_ind_y = density_point.numpy_index_y
    np_ind_z = density_point.numpy_index_z
    np_indexes = (np_ind_x, np_ind_y, np_ind_z)

    den_chi = density_point.calc_density_chi(cell, atom_site_susceptibility)
    den_f = numpy.array(density_point.density_ferro, dtype=float)
    den_a = numpy.array(density_point.density_antiferro, dtype=float)

    points_a = mem_parameters.points_a
    points_b = mem_parameters.points_b
    points_c = mem_parameters.points_c
    chi_f = mem_parameters.chi_ferro
    chi_a = mem_parameters.chi_antiferro

    n_xyz = (points_a, points_b, points_c)
    den_chi_3d = transfer_to_density_3d(np_indexes, den_chi, n_xyz, r_ij, b_i)

    den_b = chi_f*den_f + chi_a*den_a
    den_b_3d = transfer_to_density_3d(np_indexes, den_b, n_xyz, r_ij, b_i)

    fract_x, fract_y, fract_z = section.calc_fractions(cell, atom_site)
    fract_x_uc, fract_y_uc = fract_x % 1., fract_y % 1.
    fract_z_uc = fract_z % 1.

    ind_xyz = numpy.array([fract_x_uc*points_a, fract_y_uc*points_b,
                           fract_z_uc*points_c], dtype=float).transpose()

    den_chi_section = tri_linear_interpolation(ind_xyz, den_chi_3d)
    den_b_section = tri_linear_interpolation(ind_xyz, den_b_3d)

    points_x = section.points_x
    points_y = section.points_y
    den_chi_section = den_chi_section.reshape((points_x, points_y))
    den_b_section = den_b_section.reshape((points_x, points_y))
    return den_chi_section, den_b_section
